/*
    Copyright 2007, Joe Davidson <joedavidson@gmail.com>

    This file is part of FFTPatcher.

    FFTPatcher is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FFTPatcher is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FFTPatcher.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Xml;
using ICSharpCode.SharpZipLib.GZip;
using ICSharpCode.SharpZipLib.Tar;
using PatcherLib.Utilities;

namespace PatcherLib.Iso
{
    public static class FFTPack
    {
		#regionPublicMethods(7)
        static XmlNode resourcesDoc;
        public const int NumFftPackFiles = 3970;

        static FFTPack()
        {
            using( MemoryStream memStream = new MemoryStream( PatcherLib.Properties.Resources.Resources_tar, false ) )
            using( GZipInputStream gzStream = new GZipInputStream( memStream ) )
            using ( TarInputStream tarStream = new TarInputStream( gzStream ) )
            {
                TarEntry entry = tarStream.GetNextEntry();
                while ( entry != null )
                {
                    if ( entry.Name == "FFTPackFiles.xml" )
                    {
                        XmlDocument doc = new XmlDocument();
                        doc.Load( tarStream );
                        resourcesDoc = doc;
                    }
                    entry = tarStream.GetNextEntry();
                }
            }
        }
        private static Dictionary<int, string> fftPackFiles;

        public static Dictionary<int, string> FFTPackFiles
        {
            get
            {
                if (fftPackFiles == null)
                {
                    try
                    {
                        fftPackFiles = new Dictionary<int, string>();

                        XmlNodeList nodes = resourcesDoc.SelectNodes("/files/file");
                        foreach (XmlNode node in nodes)
                        {
                            int key = Convert.ToInt32(node.Attributes["entry"].InnerText);
                            if (fftPackFiles.ContainsKey(key))
                            {
                                throw new Exception();
                            }
                            fftPackFiles.Add(Convert.ToInt32(node.Attributes["entry"].InnerText), node.Attributes["name"].InnerText);
                        }
                    }
                    catch (Exception)
                    {
                        fftPackFiles = null;
                        throw;
                    }
                }

                return fftPackFiles;
            }
        }


        public static void DumpToDirectory( string filename, string path, BackgroundWorker worker )
        {
            FileStream stream = null;

            try
            {
                stream = new FileStream( filename, FileMode.Open, FileAccess.Read );
                DumpToDirectory(stream, path, stream.Length, worker);
            }
            catch( Exception )
            {
                throw;
            }
            finally
            {
                if( stream != null )
                {
                    stream.Close();
                }
            }
        }

        public static void DumpToDirectory( Stream stream, string path, long fftpackLength, BackgroundWorker worker )
        {
            MakeDirectories( path, "BATTLE", "EFFECT", "EVENT", "MAP", "MENU", "SOUND", "WORLD", "SAVEIMAGE", "unknown", "OPEN" );
            for( int i = 1; i <= NumFftPackFiles; i++ )
            {
                byte[] bytes = GetFile( stream, i, fftpackLength );
                string filename = string.Empty;

                if( FFTPackFiles.ContainsKey( i ) )
                {
                    filename = FFTPackFiles[i];
                }
                else if( bytes.Length == 0 )
                {
                    filename = string.Format( "unknown/fftpack.{0}.dummy", i );
                }
                else
                {
                    filename = string.Format( "unknown/fftpack.{0}", i );
                }

                if (worker != null)
                {
                    worker.ReportProgress((i * 100) / NumFftPackFiles, string.Format("Extracting {0}", filename));
                }

                filename = Path.Combine( path, filename );

                SaveToFile( bytes, filename );

            }
        }

        public static byte[] GetFile( Stream stream, int index, long fftPackLength )
        {
            return GetFileFromFFTPack(stream, (Files)index, fftPackLength);
        }

        public static byte[] GetFileFromIso( Stream stream, PatcherLib.Iso.PspIso.PspIsoInfo info, Files file, int start, int length )
        {
            stream.Seek( info[PspIso.Sectors.PSP_GAME_USRDIR_fftpack_bin] * 2048, SeekOrigin.Begin );
            return GetFileFromFFTPack( stream, file, info.GetFileSize( PspIso.Sectors.PSP_GAME_USRDIR_fftpack_bin ), start, length );
        }

        public static byte[] GetFileFromIso( Stream stream, PatcherLib.Iso.PspIso.PspIsoInfo info, Files file )
        {
            stream.Seek( info[PspIso.Sectors.PSP_GAME_USRDIR_fftpack_bin] * 2048, SeekOrigin.Begin );
            return GetFileFromFFTPack(stream, file, info.GetFileSize(PspIso.Sectors.PSP_GAME_USRDIR_fftpack_bin));
        }

        private static UInt32 GetStartPosition( Stream stream, Files file, long fftpackLength )
        {
            long pos = stream.Position;
            byte[] loc = new byte[4];
            int index = (int)file;
            stream.Seek( index * 4 + 4 + pos, SeekOrigin.Begin );
            stream.Read( loc, 0, 4 );
            UInt32 result = loc.ToUInt32();
            stream.Seek( pos, SeekOrigin.Begin );
            return result;
        }

        private static UInt32 GetEndPosition( Stream stream, Files file, long fftpackLength )
        {
            long pos = stream.Position;
            byte[] loc = new byte[4];
            int index = (int)file;
            UInt32 end;
            if (index >= NumFftPackFiles)
            {
                end = (uint)fftpackLength;
            }
            else
            {
                stream.Seek( index * 4 + 4 + pos + 4, SeekOrigin.Begin );
                stream.Read( loc, 0, 4 );
                end = loc.ToUInt32();
                stream.Seek( pos, SeekOrigin.Begin );
            }
            return end;
        }

        private static byte[] GetFileFromFFTPack( Stream stream, Files file, long fftpackLength, int start, int length )
        {
            long pos = stream.Position;
            UInt32 fileStart = GetStartPosition( stream, file, fftpackLength );
            
            UInt32 fileEnd = GetEndPosition( stream, file, fftpackLength );

            UInt32 fileLength = fileEnd - fileStart;
            if (start > fileLength) throw new ArgumentOutOfRangeException( "start", "start longer than file length" );
            if (length > fileLength) throw new ArgumentOutOfRangeException( "length", "length longer than file length" );

            byte[] result = new byte[length];
            stream.Seek( fileStart + start + pos, SeekOrigin.Begin );
            stream.Read( result, 0, length );
            stream.Seek( pos, SeekOrigin.Begin );

            return result;
        }

        public static byte[] GetFileFromFFTPack( Stream stream, Files file, long fftpackLength )
        {
            long pos = stream.Position;
            UInt32 start = GetStartPosition( stream, file, fftpackLength );

            UInt32 end = GetEndPosition( stream, file, fftpackLength );

            UInt32 length = end - start;
            byte[] result = new byte[length];
            stream.Seek( start + pos, SeekOrigin.Begin );
            stream.Read( result, 0, (int)length );
            stream.Seek(pos, SeekOrigin.Begin);

            return result;
        }

        public static void MergeDumpedFiles( string path, string filename, BackgroundWorker worker )
        {
            FileStream stream = null;
            try
            {
                stream = new FileStream( filename, FileMode.Create );
                stream.WriteByte( 0x80 );
                stream.WriteByte( 0x0F );
                stream.WriteByte( 0x00 );
                stream.WriteByte( 0x00 );
                stream.WriteByte( 0x10 );
                stream.WriteByte( 0x00 );
                stream.WriteByte( 0x00 );
                stream.WriteByte( 0x00 );

                stream.Seek( 15888, SeekOrigin.Begin );
                UInt32 end = (UInt32)stream.Position;
                UInt32 start = 8;

                for( int i = 1; i <= NumFftPackFiles; i++ )
                {
                    stream.Seek( start, SeekOrigin.Begin );
                    stream.Write( end.ToBytes(), 0, 4 );
                    start += 4;

                    stream.Seek( end, SeekOrigin.Begin );
                    
                    byte[] bytes = ReadFile( path, i );
                    if( bytes.Length != 0 )
                    {
                        stream.Write( bytes, 0, bytes.Length );
                    }

                    end = (UInt32)stream.Position;
                    if (worker != null)
                        worker.ReportProgress( (i * 100) / NumFftPackFiles );
                }
            }
            catch( Exception )
            {
                throw;
            }
            finally
            {
                if( stream != null )
                {
                    stream.Flush();
                    stream.Close();
                    stream = null;
                }
            }
        }

        public static void PatchFile( Stream stream, PatcherLib.Iso.PspIso.PspIsoInfo info, int index, int offset, byte[] bytes )
        {
            try
            {
                long fftpackLocation = info[PspIso.Sectors.PSP_GAME_USRDIR_fftpack_bin] * 2048;
                stream.Seek( fftpackLocation + ( index - 1 ) * 4 + 8, SeekOrigin.Begin );
                byte[] pointer = new byte[4];
                stream.Read( pointer, 0, 4 );

                stream.Seek( fftpackLocation + pointer.ToUInt32() + offset, SeekOrigin.Begin );

                stream.Write( bytes, 0, bytes.Length );
            }
            catch ( Exception )
            {
                throw;
            }
        }

		#endregionPublicMethods

		#regionPrivateMethods(3)

        private static void MakeDirectories( string path, params string[] dirs )
        {
            foreach( string dir in dirs )
            {
                Directory.CreateDirectory( Path.Combine( path, dir ) );
            }
        }

        private static byte[] ReadFile( string path, int index )
        {
            FileStream stream = null;

            try
            {
                if( FFTPackFiles.ContainsKey( index ) && File.Exists( Path.Combine( path, FFTPackFiles[index] ) ) )
                {
                    stream = new FileStream( Path.Combine( path, FFTPackFiles[index] ), FileMode.Open );
                }
                else if( File.Exists( Path.Combine( path, string.Format( "unknown/fftpack.{0}", index ) ) ) )
                {
                    stream = new FileStream( Path.Combine( path, string.Format( "unknown/fftpack.{0}", index ) ), FileMode.Open );
                }
                else if( File.Exists( Path.Combine( path, string.Format( "unknown/fftpack.{0}.dummy", index ) ) ) )
                {
                    return new byte[0];
                }
                else
                {
                    throw new Exception();
                }

                byte[] result = new byte[stream.Length];
                stream.Read( result, 0, (int)stream.Length );

                return result;
            }
            catch( Exception )
            {
                throw;
            }
            finally
            {
                if( stream != null )
                {
                    stream.Close();
                    stream = null;
                }
            }
        }

        private static void SaveToFile( byte[] bytes, string path )
        {
            FileStream stream = null;
            try
            {
                stream = new FileStream( path, FileMode.Create );
                stream.Write( bytes, 0, bytes.Length );
            }
            catch( Exception )
            {
                throw;
            }
            finally
            {
                if( stream != null )
                {
                    stream.Flush();
                    stream.Close();
                }
            }
        }

		#endregionPrivateMethods

        public enum Files
        {
            SYSTEM_CNF = 3,
            SLPS_007_70 = 4,
            BATTLE_BIN = 5,
            EVENT_SMALL_OUT = 6,
            EVENT_DEBUGMAP_OUT = 7,
            EVENT_OPTION_OUT = 8,
            EVENT_REQUIRE_OUT = 9,
            EVENT_HELPMENU_OUT = 10,
            EVENT_JOBSTTS_OUT = 11,
            EVENT_ATTACK_OUT = 12,
            EVENT_DEBUGCHR_OUT = 13,
            EVENT_ETC_OUT = 14,
            EVENT_EQUIP_OUT = 15,
            EVENT_CARD_OUT = 16,
            EVENT_BUNIT_OUT = 17,
            EVENT_EVTOOL_OUT = 18,
            EVENT_MAPTITLE_BIN = 19,
            EVENT_FONT_BIN = 20,
            EVENT_FRAME_BIN = 21,
            EVENT_TEST_EVT = 22,
            EVENT_EVTFACE_BIN = 23,
            EVENT_UNIT_BIN = 24,
            EVENT_BTLEVT_BIN = 25,
            EVENT_CHAPTER1_OUT = 26,
            EVENT_CHAPTER2_OUT = 27,
            EVENT_CHAPTER3_OUT = 28,
            EVENT_CHAPTER4_OUT = 29,
            EVENT_GAMEOVER_BIN = 30,
            EVENT_BONUS_BIN = 31,
            EVENT_WIN001_BIN = 32,
            EVENT_WLDFACE4_BIN = 34,
            EVENT_WLDFACE_BIN = 35,
            EVENT_END1_BIN = 36,
            EVENT_END2_BIN = 37,
            EVENT_END3_BIN = 38,
            EVENT_END4_BIN = 39,
            EVENT_END5_BIN = 40,
            EVENT_ATCHELP_LZW = 41,
            EVENT_SPELL_MES = 42,
            EVENT_SAMPLE_LZW = 43,
            EVENT_WORLD_LZW = 44,
            EVENT_OPEN_LZW = 45,
            EVENT_JOIN_LZW = 46,
            EVENT_HELP_LZW = 47,
            EVENT_WLDHELP_LZW = 48,
            EVENT_EVTCHR_BIN = 49,
            BATTLE_EFC_FNT_TIM = 50,
            BATTLE_TYPE1_SHP = 51,
            BATTLE_TYPE2_SHP = 52,
            BATTLE_CYOKO_SHP = 53,
            BATTLE_MON_SHP = 54,
            BATTLE_OTHER_SHP = 55,
            BATTLE_ARUTE_SHP = 56,
            BATTLE_KANZEN_SHP = 57,
            BATTLE_TYPE1_SEQ = 58,
            BATTLE_TYPE2_SEQ = 59,
            BATTLE_TYPE3_SEQ = 60,
            BATTLE_TYPE4_SEQ = 61,
            BATTLE_CYOKO_SEQ = 62,
            BATTLE_MON_SEQ = 63,
            BATTLE_OTHER_SEQ = 64,
            BATTLE_RUKA_SEQ = 65,
            BATTLE_ARUTE_SEQ = 66,
            BATTLE_KANZEN_SEQ = 67,
            BATTLE_WEP1_SHP = 68,
            BATTLE_WEP2_SHP = 69,
            BATTLE_WEP1_SEQ = 70,
            BATTLE_WEP2_SEQ = 71,
            BATTLE_EFF1_SHP = 72,
            BATTLE_EFF2_SHP = 73,
            BATTLE_EFF1_SEQ = 74,
            BATTLE_EFF2_SEQ = 75,
            BATTLE_WEP_SPR = 76,
            BATTLE_OTHER_SPR = 77,
            BATTLE_ADORA_SPR = 78,
            BATTLE_AGURI_SPR = 79,
            BATTLE_AJORA_SPR = 80,
            BATTLE_ARLI_SPR = 81,
            BATTLE_ARU_SPR = 82,
            BATTLE_ARUFU_SPR = 83,
            BATTLE_ARUMA_SPR = 84,
            BATTLE_ARUTE_SPR = 85,
            BATTLE_BARITEN_SPR = 86,
            BATTLE_BARU_SPR = 87,
            BATTLE_BARUNA_SPR = 88,
            BATTLE_BEHI_SPR = 89,
            BATTLE_BEIO_SPR = 90,
            BATTLE_BIBUROS_SPR = 91,
            BATTLE_BOM_SPR = 92,
            BATTLE_CLOUD_SPR = 93,
            BATTLE_CYOKO_SPR = 94,
            BATTLE_DAISU_SPR = 95,
            BATTLE_DAMI_SPR = 96,
            BATTLE_DEMON_SPR = 97,
            BATTLE_DILY_SPR = 98,
            BATTLE_DILY2_SPR = 99,
            BATTLE_DILY3_SPR = 100,
            BATTLE_DORA_SPR = 101,
            BATTLE_DORA1_SPR = 102,
            BATTLE_DORA2_SPR = 103,
            BATTLE_ERU_SPR = 104,
            BATTLE_FURAIA_SPR = 105,
            BATTLE_FUSUI_M_SPR = 106,
            BATTLE_FUSUI_W_SPR = 107,
            BATTLE_FYUNE_SPR = 108,
            BATTLE_MARA_SPR = 109,
            BATTLE_MUSU_SPR = 110,
            BATTLE_GIN_M_SPR = 111,
            BATTLE_GOB_SPR = 112,
            BATTLE_GORU_SPR = 113,
            BATTLE_GYUMU_SPR = 114,
            BATTLE_H61_SPR = 115,
            BATTLE_H75_SPR = 116,
            BATTLE_H76_SPR = 117,
            BATTLE_H77_SPR = 118,
            BATTLE_H78_SPR = 119,
            BATTLE_RAFA_SPR = 120,
            BATTLE_H80_SPR = 121,
            BATTLE_H81_SPR = 122,
            BATTLE_H82_SPR = 123,
            BATTLE_H83_SPR = 124,
            BATTLE_H85_SPR = 125,
            BATTLE_HASYU_SPR = 126,
            BATTLE_HEBI_SPR = 127,
            BATTLE_HIME_SPR = 128,
            BATTLE_HYOU_SPR = 129,
            BATTLE_IKA_SPR = 130,
            BATTLE_ITEM_M_SPR = 131,
            BATTLE_ITEM_W_SPR = 132,
            BATTLE_KANBA_SPR = 133,
            BATTLE_KANZEN_SPR = 134,
            BATTLE_KASANEK_SPR = 135,
            BATTLE_KASANEM_SPR = 136,
            BATTLE_KI_SPR = 137,
            BATTLE_KNIGHT_M_SPR = 138,
            BATTLE_KNIGHT_W_SPR = 139,
            BATTLE_KURO_M_SPR = 140,
            BATTLE_KURO_W_SPR = 141,
            BATTLE_KYUKU_SPR = 142,
            BATTLE_LEDY_SPR = 143,
            BATTLE_GANDO_SPR = 144,
            BATTLE_MINA_M_SPR = 145,
            BATTLE_MINA_W_SPR = 146,
            BATTLE_MINOTA_SPR = 147,
            BATTLE_MOL_SPR = 148,
            BATTLE_MONK_M_SPR = 149,
            BATTLE_MONK_W_SPR = 150,
            BATTLE_MONO_M_SPR = 151,
            BATTLE_MONO_W_SPR = 152,
            BATTLE_GARU_SPR = 153,
            BATTLE_NINJA_M_SPR = 154,
            BATTLE_NINJA_W_SPR = 155,
            BATTLE_ODORI_W_SPR = 156,
            BATTLE_ONMYO_M_SPR = 157,
            BATTLE_ONMYO_W_SPR = 158,
            BATTLE_ORAN_SPR = 159,
            BATTLE_ORU_SPR = 160,
            BATTLE_H79_SPR = 161,
            BATTLE_RAGU_SPR = 162,
            BATTLE_RAMUZA_SPR = 163,
            BATTLE_RAMUZA2_SPR = 164,
            BATTLE_RAMUZA3_SPR = 165,
            BATTLE_REZE_SPR = 166,
            BATTLE_REZE_D_SPR = 167,
            BATTLE_RUDO_SPR = 168,
            BATTLE_RYU_M_SPR = 169,
            BATTLE_RYU_W_SPR = 170,
            BATTLE_SAMU_M_SPR = 171,
            BATTLE_SAMU_W_SPR = 172,
            BATTLE_SAN_M_SPR = 173,
            BATTLE_SAN_W_SPR = 174,
            BATTLE_SERIA_SPR = 175,
            BATTLE_SIMON_SPR = 176,
            BATTLE_SIRO_M_SPR = 177,
            BATTLE_SIRO_W_SPR = 178,
            BATTLE_SUKERU_SPR = 179,
            BATTLE_SYOU_M_SPR = 180,
            BATTLE_SYOU_W_SPR = 181,
            BATTLE_TETSU_SPR = 182,
            BATTLE_THIEF_M_SPR = 183,
            BATTLE_THIEF_W_SPR = 184,
            BATTLE_TOKI_M_SPR = 185,
            BATTLE_TOKI_W_SPR = 186,
            BATTLE_TORI_SPR = 187,
            BATTLE_URI_SPR = 188,
            BATTLE_VERI_SPR = 189,
            BATTLE_VORU_SPR = 190,
            BATTLE_WAJU_M_SPR = 191,
            BATTLE_WAJU_W_SPR = 192,
            BATTLE_WIGU_SPR = 193,
            BATTLE_YUMI_M_SPR = 194,
            BATTLE_YUMI_W_SPR = 195,
            BATTLE_YUREI_SPR = 196,
            BATTLE_ZARU_SPR = 197,
            BATTLE_ZARU2_SPR = 198,
            BATTLE_ZARUE_SPR = 199,
            BATTLE_ZARUMOU_SPR = 200,
            BATTLE_10M_SPR = 201,
            BATTLE_10W_SPR = 202,
            BATTLE_20M_SPR = 203,
            BATTLE_20W_SPR = 204,
            BATTLE_40M_SPR = 205,
            BATTLE_40W_SPR = 206,
            BATTLE_60M_SPR = 207,
            BATTLE_60W_SPR = 208,
            BATTLE_CYOMON1_SPR = 209,
            BATTLE_CYOMON2_SPR = 210,
            BATTLE_CYOMON3_SPR = 211,
            BATTLE_CYOMON4_SPR = 212,
            BATTLE_SOURYO_SPR = 213,
            BATTLE_ARLI2_SP2 = 214,
            BATTLE_BEHI2_SP2 = 215,
            BATTLE_BIBU2_SP2 = 216,
            BATTLE_BOM2_SP2 = 217,
            BATTLE_DEMON2_SP2 = 218,
            BATTLE_DORA22_SP2 = 219,
            BATTLE_HYOU2_SP2 = 220,
            BATTLE_IRON2_SP2 = 221,
            BATTLE_IRON3_SP2 = 222,
            BATTLE_IRON4_SP2 = 223,
            BATTLE_IRON5_SP2 = 224,
            BATTLE_MINOTA2_SP2 = 225,
            BATTLE_MOL2_SP2 = 226,
            BATTLE_TORI2_SP2 = 227,
            BATTLE_URI2_SP2 = 228,
            BATTLE_ENTD1_ENT = 229,
            BATTLE_ENTD2_ENT = 230,
            BATTLE_ENTD3_ENT = 231,
            BATTLE_ENTD4_ENT = 232,
            BATTLE_ZODIAC_BIN = 233,
            EFFECT_E000_BIN = 234,
            EFFECT_E001_BIN = 235,
            EFFECT_E002_BIN = 236,
            EFFECT_E003_BIN = 237,
            EFFECT_E004_BIN = 238,
            EFFECT_E005_BIN = 239,
            EFFECT_E006_BIN = 240,
            EFFECT_E007_BIN = 241,
            EFFECT_E008_BIN = 242,
            EFFECT_E009_BIN = 243,
            EFFECT_E010_BIN = 244,
            EFFECT_E011_BIN = 245,
            EFFECT_E012_BIN = 246,
            EFFECT_E013_BIN = 247,
            EFFECT_E014_BIN = 248,
            EFFECT_E015_BIN = 249,
            EFFECT_E016_BIN = 250,
            EFFECT_E017_BIN = 251,
            EFFECT_E018_BIN = 252,
            EFFECT_E019_BIN = 253,
            EFFECT_E020_BIN = 254,
            EFFECT_E021_BIN = 255,
            EFFECT_E022_BIN = 256,
            EFFECT_E023_BIN = 257,
            EFFECT_E024_BIN = 258,
            EFFECT_E025_BIN = 259,
            EFFECT_E026_BIN = 260,
            EFFECT_E027_BIN = 261,
            EFFECT_E028_BIN = 262,
            EFFECT_E029_BIN = 263,
            EFFECT_E030_BIN = 264,
            EFFECT_E031_BIN = 265,
            EFFECT_E032_BIN = 266,
            EFFECT_E033_BIN = 267,
            EFFECT_E034_BIN = 268,
            EFFECT_E035_BIN = 269,
            EFFECT_E036_BIN = 270,
            EFFECT_E037_BIN = 271,
            EFFECT_E038_BIN = 272,
            EFFECT_E039_BIN = 273,
            EFFECT_E040_BIN = 274,
            EFFECT_E041_BIN = 275,
            EFFECT_E042_BIN = 276,
            EFFECT_E043_BIN = 277,
            EFFECT_E044_BIN = 278,
            EFFECT_E045_BIN = 279,
            EFFECT_E046_BIN = 280,
            EFFECT_E047_BIN = 281,
            EFFECT_E048_BIN = 282,
            EFFECT_E049_BIN = 283,
            EFFECT_E050_BIN = 284,
            EFFECT_E051_BIN = 285,
            EFFECT_E052_BIN = 286,
            EFFECT_E053_BIN = 287,
            EFFECT_E054_BIN = 288,
            EFFECT_E055_BIN = 289,
            EFFECT_E056_BIN = 290,
            EFFECT_E057_BIN = 291,
            EFFECT_E058_BIN = 292,
            EFFECT_E059_BIN = 293,
            EFFECT_E060_BIN = 294,
            EFFECT_E061_BIN = 295,
            EFFECT_E062_BIN = 296,
            EFFECT_E063_BIN = 297,
            EFFECT_E064_BIN = 298,
            EFFECT_E065_BIN = 299,
            EFFECT_E066_BIN = 300,
            EFFECT_E067_BIN = 301,
            EFFECT_E068_BIN = 302,
            EFFECT_E069_BIN = 303,
            EFFECT_E070_BIN = 304,
            EFFECT_E071_BIN = 305,
            EFFECT_E072_BIN = 306,
            EFFECT_E073_BIN = 307,
            EFFECT_E074_BIN = 308,
            EFFECT_E075_BIN = 309,
            EFFECT_E076_BIN = 310,
            EFFECT_E077_BIN = 311,
            EFFECT_E078_BIN = 312,
            EFFECT_E079_BIN = 313,
            EFFECT_E080_BIN = 314,
            EFFECT_E081_BIN = 315,
            EFFECT_E082_BIN = 316,
            EFFECT_E083_BIN = 317,
            EFFECT_E084_BIN = 318,
            EFFECT_E085_BIN = 319,
            EFFECT_E086_BIN = 320,
            EFFECT_E087_BIN = 321,
            EFFECT_E088_BIN = 322,
            EFFECT_E089_BIN = 323,
            EFFECT_E090_BIN = 324,
            EFFECT_E091_BIN = 325,
            EFFECT_E092_BIN = 326,
            EFFECT_E093_BIN = 327,
            EFFECT_E094_BIN = 328,
            EFFECT_E095_BIN = 329,
            EFFECT_E096_BIN = 330,
            EFFECT_E097_BIN = 331,
            EFFECT_E098_BIN = 332,
            EFFECT_E099_BIN = 333,
            EFFECT_E100_BIN = 334,
            EFFECT_E101_BIN = 335,
            EFFECT_E102_BIN = 336,
            EFFECT_E103_BIN = 337,
            EFFECT_E104_BIN = 338,
            EFFECT_E105_BIN = 339,
            EFFECT_E106_BIN = 340,
            EFFECT_E107_BIN = 341,
            EFFECT_E108_BIN = 342,
            EFFECT_E109_BIN = 343,
            EFFECT_E110_BIN = 344,
            EFFECT_E111_BIN = 345,
            EFFECT_E112_BIN = 346,
            EFFECT_E113_BIN = 347,
            EFFECT_E114_BIN = 348,
            EFFECT_E115_BIN = 349,
            EFFECT_E116_BIN = 350,
            EFFECT_E117_BIN = 351,
            EFFECT_E118_BIN = 352,
            EFFECT_E119_BIN = 353,
            EFFECT_E120_BIN = 354,
            EFFECT_E121_BIN = 355,
            EFFECT_E122_BIN = 356,
            EFFECT_E123_BIN = 357,
            EFFECT_E124_BIN = 358,
            EFFECT_E125_BIN = 359,
            EFFECT_E126_BIN = 360,
            EFFECT_E127_BIN = 361,
            EFFECT_E128_BIN = 362,
            EFFECT_E129_BIN = 363,
            EFFECT_E130_BIN = 364,
            EFFECT_E131_BIN = 365,
            EFFECT_E132_BIN = 366,
            EFFECT_E133_BIN = 367,
            EFFECT_E134_BIN = 368,
            EFFECT_E135_BIN = 369,
            EFFECT_E136_BIN = 370,
            EFFECT_E137_BIN = 371,
            EFFECT_E138_BIN = 372,
            EFFECT_E139_BIN = 373,
            EFFECT_E140_BIN = 374,
            EFFECT_E141_BIN = 375,
            EFFECT_E142_BIN = 376,
            EFFECT_E143_BIN = 377,
            EFFECT_E144_BIN = 378,
            EFFECT_E145_BIN = 379,
            EFFECT_E146_BIN = 380,
            EFFECT_E147_BIN = 381,
            EFFECT_E148_BIN = 382,
            EFFECT_E149_BIN = 383,
            EFFECT_E150_BIN = 384,
            EFFECT_E151_BIN = 385,
            EFFECT_E152_BIN = 386,
            EFFECT_E153_BIN = 387,
            EFFECT_E154_BIN = 388,
            EFFECT_E155_BIN = 389,
            EFFECT_E156_BIN = 390,
            EFFECT_E157_BIN = 391,
            EFFECT_E158_BIN = 392,
            EFFECT_E159_BIN = 393,
            EFFECT_E160_BIN = 394,
            EFFECT_E161_BIN = 395,
            EFFECT_E162_BIN = 396,
            EFFECT_E163_BIN = 397,
            EFFECT_E164_BIN = 398,
            EFFECT_E165_BIN = 399,
            EFFECT_E166_BIN = 400,
            EFFECT_E167_BIN = 401,
            EFFECT_E168_BIN = 402,
            EFFECT_E169_BIN = 403,
            EFFECT_E170_BIN = 404,
            EFFECT_E171_BIN = 405,
            EFFECT_E172_BIN = 406,
            EFFECT_E173_BIN = 407,
            EFFECT_E174_BIN = 408,
            EFFECT_E175_BIN = 409,
            EFFECT_E176_BIN = 410,
            EFFECT_E177_BIN = 411,
            EFFECT_E178_BIN = 412,
            EFFECT_E179_BIN = 413,
            EFFECT_E180_BIN = 414,
            EFFECT_E181_BIN = 415,
            EFFECT_E182_BIN = 416,
            EFFECT_E183_BIN = 417,
            EFFECT_E184_BIN = 418,
            EFFECT_E185_BIN = 419,
            EFFECT_E186_BIN = 420,
            EFFECT_E187_BIN = 421,
            EFFECT_E188_BIN = 422,
            EFFECT_E189_BIN = 423,
            EFFECT_E190_BIN = 424,
            EFFECT_E191_BIN = 425,
            EFFECT_E192_BIN = 426,
            EFFECT_E193_BIN = 427,
            EFFECT_E194_BIN = 428,
            EFFECT_E195_BIN = 429,
            EFFECT_E196_BIN = 430,
            EFFECT_E197_BIN = 431,
            EFFECT_E198_BIN = 432,
            EFFECT_E199_BIN = 433,
            EFFECT_E200_BIN = 434,
            EFFECT_E201_BIN = 435,
            EFFECT_E202_BIN = 436,
            EFFECT_E203_BIN = 437,
            EFFECT_E204_BIN = 438,
            EFFECT_E205_BIN = 439,
            EFFECT_E206_BIN = 440,
            EFFECT_E207_BIN = 441,
            EFFECT_E208_BIN = 442,
            EFFECT_E209_BIN = 443,
            EFFECT_E210_BIN = 444,
            EFFECT_E211_BIN = 445,
            EFFECT_E212_BIN = 446,
            EFFECT_E213_BIN = 447,
            EFFECT_E214_BIN = 448,
            EFFECT_E215_BIN = 449,
            EFFECT_E216_BIN = 450,
            EFFECT_E217_BIN = 451,
            EFFECT_E218_BIN = 452,
            EFFECT_E219_BIN = 453,
            EFFECT_E220_BIN = 454,
            EFFECT_E221_BIN = 455,
            EFFECT_E222_BIN = 456,
            EFFECT_E223_BIN = 457,
            EFFECT_E224_BIN = 458,
            EFFECT_E225_BIN = 459,
            EFFECT_E226_BIN = 460,
            EFFECT_E227_BIN = 461,
            EFFECT_E228_BIN = 462,
            EFFECT_E229_BIN = 463,
            EFFECT_E230_BIN = 464,
            EFFECT_E231_BIN = 465,
            EFFECT_E232_BIN = 466,
            EFFECT_E233_BIN = 467,
            EFFECT_E234_BIN = 468,
            EFFECT_E235_BIN = 469,
            EFFECT_E236_BIN = 470,
            EFFECT_E237_BIN = 471,
            EFFECT_E238_BIN = 472,
            EFFECT_E239_BIN = 473,
            EFFECT_E240_BIN = 474,
            EFFECT_E241_BIN = 475,
            EFFECT_E242_BIN = 476,
            EFFECT_E243_BIN = 477,
            EFFECT_E244_BIN = 478,
            EFFECT_E245_BIN = 479,
            EFFECT_E246_BIN = 480,
            EFFECT_E247_BIN = 481,
            EFFECT_E248_BIN = 482,
            EFFECT_E249_BIN = 483,
            EFFECT_E250_BIN = 484,
            EFFECT_E251_BIN = 485,
            EFFECT_E252_BIN = 486,
            EFFECT_E253_BIN = 487,
            EFFECT_E254_BIN = 488,
            EFFECT_E255_BIN = 489,
            EFFECT_E256_BIN = 490,
            EFFECT_E257_BIN = 491,
            EFFECT_E258_BIN = 492,
            EFFECT_E259_BIN = 493,
            EFFECT_E260_BIN = 494,
            EFFECT_E261_BIN = 495,
            EFFECT_E262_BIN = 496,
            EFFECT_E263_BIN = 497,
            EFFECT_E264_BIN = 498,
            EFFECT_E265_BIN = 499,
            EFFECT_E266_BIN = 500,
            EFFECT_E267_BIN = 501,
            EFFECT_E268_BIN = 502,
            EFFECT_E269_BIN = 503,
            EFFECT_E270_BIN = 504,
            EFFECT_E271_BIN = 505,
            EFFECT_E272_BIN = 506,
            EFFECT_E273_BIN = 507,
            EFFECT_E274_BIN = 508,
            EFFECT_E275_BIN = 509,
            EFFECT_E276_BIN = 510,
            EFFECT_E277_BIN = 511,
            EFFECT_E278_BIN = 512,
            EFFECT_E279_BIN = 513,
            EFFECT_E280_BIN = 514,
            EFFECT_E281_BIN = 515,
            EFFECT_E282_BIN = 516,
            EFFECT_E283_BIN = 517,
            EFFECT_E284_BIN = 518,
            EFFECT_E285_BIN = 519,
            EFFECT_E286_BIN = 520,
            EFFECT_E287_BIN = 521,
            EFFECT_E288_BIN = 522,
            EFFECT_E289_BIN = 523,
            EFFECT_E290_BIN = 524,
            EFFECT_E291_BIN = 525,
            EFFECT_E292_BIN = 526,
            EFFECT_E293_BIN = 527,
            EFFECT_E294_BIN = 528,
            EFFECT_E295_BIN = 529,
            EFFECT_E296_BIN = 530,
            EFFECT_E297_BIN = 531,
            EFFECT_E298_BIN = 532,
            EFFECT_E299_BIN = 533,
            EFFECT_E300_BIN = 534,
            EFFECT_E301_BIN = 535,
            EFFECT_E302_BIN = 536,
            EFFECT_E303_BIN = 537,
            EFFECT_E304_BIN = 538,
            EFFECT_E305_BIN = 539,
            EFFECT_E306_BIN = 540,
            EFFECT_E307_BIN = 541,
            EFFECT_E308_BIN = 542,
            EFFECT_E309_BIN = 543,
            EFFECT_E310_BIN = 544,
            EFFECT_E311_BIN = 545,
            EFFECT_E312_BIN = 546,
            EFFECT_E313_BIN = 547,
            EFFECT_E314_BIN = 548,
            EFFECT_E315_BIN = 549,
            EFFECT_E316_BIN = 550,
            EFFECT_E317_BIN = 551,
            EFFECT_E318_BIN = 552,
            EFFECT_E319_BIN = 553,
            EFFECT_E320_BIN = 554,
            EFFECT_E321_BIN = 555,
            EFFECT_E322_BIN = 556,
            EFFECT_E323_BIN = 557,
            EFFECT_E324_BIN = 558,
            EFFECT_E325_BIN = 559,
            EFFECT_E326_BIN = 560,
            EFFECT_E327_BIN = 561,
            EFFECT_E328_BIN = 562,
            EFFECT_E329_BIN = 563,
            EFFECT_E330_BIN = 564,
            EFFECT_E331_BIN = 565,
            EFFECT_E332_BIN = 566,
            EFFECT_E333_BIN = 567,
            EFFECT_E334_BIN = 568,
            EFFECT_E335_BIN = 569,
            EFFECT_E336_BIN = 570,
            EFFECT_E337_BIN = 571,
            EFFECT_E338_BIN = 572,
            EFFECT_E339_BIN = 573,
            EFFECT_E340_BIN = 574,
            EFFECT_E341_BIN = 575,
            EFFECT_E342_BIN = 576,
            EFFECT_E343_BIN = 577,
            EFFECT_E344_BIN = 578,
            EFFECT_E345_BIN = 579,
            EFFECT_E346_BIN = 580,
            EFFECT_E347_BIN = 581,
            EFFECT_E348_BIN = 582,
            EFFECT_E349_BIN = 583,
            EFFECT_E350_BIN = 584,
            EFFECT_E351_BIN = 585,
            EFFECT_E352_BIN = 586,
            EFFECT_E353_BIN = 587,
            EFFECT_E354_BIN = 588,
            EFFECT_E355_BIN = 589,
            EFFECT_E356_BIN = 590,
            EFFECT_E357_BIN = 591,
            EFFECT_E358_BIN = 592,
            EFFECT_E359_BIN = 593,
            EFFECT_E360_BIN = 594,
            EFFECT_E361_BIN = 595,
            EFFECT_E362_BIN = 596,
            EFFECT_E363_BIN = 597,
            EFFECT_E364_BIN = 598,
            EFFECT_E365_BIN = 599,
            EFFECT_E366_BIN = 600,
            EFFECT_E367_BIN = 601,
            EFFECT_E368_BIN = 602,
            EFFECT_E369_BIN = 603,
            EFFECT_E370_BIN = 604,
            EFFECT_E371_BIN = 605,
            EFFECT_E372_BIN = 606,
            EFFECT_E373_BIN = 607,
            EFFECT_E374_BIN = 608,
            EFFECT_E375_BIN = 609,
            EFFECT_E376_BIN = 610,
            EFFECT_E377_BIN = 611,
            EFFECT_E378_BIN = 612,
            EFFECT_E379_BIN = 613,
            EFFECT_E380_BIN = 614,
            EFFECT_E381_BIN = 615,
            EFFECT_E382_BIN = 616,
            EFFECT_E383_BIN = 617,
            EFFECT_E384_BIN = 618,
            EFFECT_E385_BIN = 619,
            EFFECT_E386_BIN = 620,
            EFFECT_E387_BIN = 621,
            EFFECT_E388_BIN = 622,
            EFFECT_E389_BIN = 623,
            EFFECT_E390_BIN = 624,
            EFFECT_E391_BIN = 625,
            EFFECT_E392_BIN = 626,
            EFFECT_E393_BIN = 627,
            EFFECT_E394_BIN = 628,
            EFFECT_E395_BIN = 629,
            EFFECT_E396_BIN = 630,
            EFFECT_E397_BIN = 631,
            EFFECT_E398_BIN = 632,
            EFFECT_E399_BIN = 633,
            EFFECT_E400_BIN = 634,
            EFFECT_E401_BIN = 635,
            EFFECT_E402_BIN = 636,
            EFFECT_E403_BIN = 637,
            EFFECT_E404_BIN = 638,
            EFFECT_E405_BIN = 639,
            EFFECT_E406_BIN = 640,
            EFFECT_E407_BIN = 641,
            EFFECT_E408_BIN = 642,
            EFFECT_E409_BIN = 643,
            EFFECT_E410_BIN = 644,
            EFFECT_E411_BIN = 645,
            EFFECT_E412_BIN = 646,
            EFFECT_E413_BIN = 647,
            EFFECT_E414_BIN = 648,
            EFFECT_E415_BIN = 649,
            EFFECT_E416_BIN = 650,
            EFFECT_E417_BIN = 651,
            EFFECT_E418_BIN = 652,
            EFFECT_E419_BIN = 653,
            EFFECT_E420_BIN = 654,
            EFFECT_E421_BIN = 655,
            EFFECT_E422_BIN = 656,
            EFFECT_E423_BIN = 657,
            EFFECT_E424_BIN = 658,
            EFFECT_E425_BIN = 659,
            EFFECT_E426_BIN = 660,
            EFFECT_E427_BIN = 661,
            EFFECT_E428_BIN = 662,
            EFFECT_E429_BIN = 663,
            EFFECT_E430_BIN = 664,
            EFFECT_E431_BIN = 665,
            EFFECT_E432_BIN = 666,
            EFFECT_E433_BIN = 667,
            EFFECT_E434_BIN = 668,
            EFFECT_E435_BIN = 669,
            EFFECT_E436_BIN = 670,
            EFFECT_E437_BIN = 671,
            EFFECT_E438_BIN = 672,
            EFFECT_E439_BIN = 673,
            EFFECT_E440_BIN = 674,
            EFFECT_E441_BIN = 675,
            EFFECT_E442_BIN = 676,
            EFFECT_E443_BIN = 677,
            EFFECT_E444_BIN = 678,
            EFFECT_E445_BIN = 679,
            EFFECT_E446_BIN = 680,
            EFFECT_E447_BIN = 681,
            EFFECT_E448_BIN = 682,
            EFFECT_E449_BIN = 683,
            EFFECT_E450_BIN = 684,
            EFFECT_E451_BIN = 685,
            EFFECT_E452_BIN = 686,
            EFFECT_E453_BIN = 687,
            EFFECT_E454_BIN = 688,
            EFFECT_E455_BIN = 689,
            EFFECT_E456_BIN = 690,
            EFFECT_E457_BIN = 691,
            EFFECT_E458_BIN = 692,
            EFFECT_E459_BIN = 693,
            EFFECT_E460_BIN = 694,
            EFFECT_E461_BIN = 695,
            EFFECT_E462_BIN = 696,
            EFFECT_E463_BIN = 697,
            EFFECT_E464_BIN = 698,
            EFFECT_E465_BIN = 699,
            EFFECT_E466_BIN = 700,
            EFFECT_E467_BIN = 701,
            EFFECT_E468_BIN = 702,
            EFFECT_E469_BIN = 703,
            EFFECT_E470_BIN = 704,
            EFFECT_E471_BIN = 705,
            EFFECT_E472_BIN = 706,
            EFFECT_E473_BIN = 707,
            EFFECT_E474_BIN = 708,
            EFFECT_E475_BIN = 709,
            EFFECT_E476_BIN = 710,
            EFFECT_E477_BIN = 711,
            EFFECT_E478_BIN = 712,
            EFFECT_E479_BIN = 713,
            EFFECT_E480_BIN = 714,
            EFFECT_E481_BIN = 715,
            EFFECT_E482_BIN = 716,
            EFFECT_E483_BIN = 717,
            EFFECT_E484_BIN = 718,
            EFFECT_E485_BIN = 719,
            EFFECT_E486_BIN = 720,
            EFFECT_E487_BIN = 721,
            EFFECT_E488_BIN = 722,
            EFFECT_E489_BIN = 723,
            EFFECT_E490_BIN = 724,
            EFFECT_E491_BIN = 725,
            EFFECT_E492_BIN = 726,
            EFFECT_E493_BIN = 727,
            EFFECT_E494_BIN = 728,
            EFFECT_E495_BIN = 729,
            EFFECT_E496_BIN = 730,
            EFFECT_E497_BIN = 731,
            EFFECT_E498_BIN = 732,
            EFFECT_E499_BIN = 733,
            EFFECT_E500_BIN = 734,
            EFFECT_E501_BIN = 735,
            EFFECT_E502_BIN = 736,
            EFFECT_E503_BIN = 737,
            EFFECT_E504_BIN = 738,
            EFFECT_E505_BIN = 739,
            EFFECT_E506_BIN = 740,
            EFFECT_E507_BIN = 741,
            EFFECT_E508_BIN = 742,
            EFFECT_E509_BIN = 743,
            EFFECT_E510_BIN = 744,
            EFFECT_E511_BIN = 745,
            MENU_BK_SHOP_TIM = 746,
            MENU_BK_SHOP2_TIM = 747,
            MENU_BK_SHOP3_TIM = 748,
            MENU_BK_HONE_TIM = 749,
            MENU_BK_HONE2_TIM = 750,
            MENU_BK_HONE3_TIM = 751,
            MENU_BK_FITR_TIM = 752,
            MENU_BK_FITR2_TIM = 753,
            MENU_BK_FITR3_TIM = 754,
            MENU_TUTO1_SCR = 755,
            MENU_TUTO1_MES = 756,
            MENU_TUTO2_SCR = 757,
            MENU_TUTO2_MES = 758,
            MENU_TUTO3_SCR = 759,
            MENU_TUTO3_MES = 760,
            MENU_TUTO4_SCR = 761,
            MENU_TUTO4_MES = 762,
            MENU_TUTO5_SCR = 763,
            MENU_TUTO5_MES = 764,
            MENU_TUTO6_SCR = 765,
            MENU_TUTO6_MES = 766,
            MENU_TUTO7_SCR = 767,
            MENU_TUTO7_MES = 768,
            MENU_FFTSAVE_DAT = 769,
            WORLD_WLDPIC_BIN = 771,
            WORLD_WLDMES_BIN = 772,
            WORLD_SNPLBIN_BIN = 773,
            WORLD_SNPLMES_BIN = 774,
            WORLD_WLDBK_BIN = 775,
            WORLD_WLDCORE_BIN = 776,
            WORLD_WORLD_BIN = 777,
            SOUND_ENV_SED = 778,
            SOUND_SYSTEM_SED = 779,
            SOUND_WAVESET_WD = 780,
            SOUND_MUSIC_00_SMD = 781,
            SOUND_MUSIC_01_SMD = 782,
            SOUND_MUSIC_02_SMD = 783,
            SOUND_MUSIC_03_SMD = 784,
            SOUND_MUSIC_04_SMD = 785,
            SOUND_MUSIC_05_SMD = 786,
            SOUND_MUSIC_06_SMD = 787,
            SOUND_MUSIC_07_SMD = 788,
            SOUND_MUSIC_08_SMD = 789,
            SOUND_MUSIC_09_SMD = 790,
            SOUND_MUSIC_10_SMD = 791,
            SOUND_MUSIC_11_SMD = 792,
            SOUND_MUSIC_12_SMD = 793,
            SOUND_MUSIC_13_SMD = 794,
            SOUND_MUSIC_14_SMD = 795,
            SOUND_MUSIC_15_SMD = 796,
            SOUND_MUSIC_16_SMD = 797,
            SOUND_MUSIC_17_SMD = 798,
            SOUND_MUSIC_18_SMD = 799,
            SOUND_MUSIC_19_SMD = 800,
            SOUND_MUSIC_20_SMD = 801,
            SOUND_MUSIC_21_SMD = 802,
            SOUND_MUSIC_22_SMD = 803,
            SOUND_MUSIC_23_SMD = 804,
            SOUND_MUSIC_24_SMD = 805,
            SOUND_MUSIC_25_SMD = 806,
            SOUND_MUSIC_26_SMD = 807,
            SOUND_MUSIC_27_SMD = 808,
            SOUND_MUSIC_28_SMD = 809,
            SOUND_MUSIC_29_SMD = 810,
            SOUND_MUSIC_30_SMD = 811,
            SOUND_MUSIC_31_SMD = 812,
            SOUND_MUSIC_32_SMD = 813,
            SOUND_MUSIC_33_SMD = 814,
            SOUND_MUSIC_34_SMD = 815,
            SOUND_MUSIC_35_SMD = 816,
            SOUND_MUSIC_36_SMD = 817,
            SOUND_MUSIC_37_SMD = 818,
            SOUND_MUSIC_38_SMD = 819,
            SOUND_MUSIC_39_SMD = 820,
            SOUND_MUSIC_40_SMD = 821,
            SOUND_MUSIC_41_SMD = 822,
            SOUND_MUSIC_42_SMD = 823,
            SOUND_MUSIC_43_SMD = 824,
            SOUND_MUSIC_44_SMD = 825,
            SOUND_MUSIC_45_SMD = 826,
            SOUND_MUSIC_46_SMD = 827,
            SOUND_MUSIC_47_SMD = 828,
            SOUND_MUSIC_48_SMD = 829,
            SOUND_MUSIC_49_SMD = 830,
            SOUND_MUSIC_50_SMD = 831,
            SOUND_MUSIC_51_SMD = 832,
            SOUND_MUSIC_52_SMD = 833,
            SOUND_MUSIC_53_SMD = 834,
            SOUND_MUSIC_54_SMD = 835,
            SOUND_MUSIC_55_SMD = 836,
            SOUND_MUSIC_56_SMD = 837,
            SOUND_MUSIC_57_SMD = 838,
            SOUND_MUSIC_58_SMD = 839,
            SOUND_MUSIC_59_SMD = 840,
            SOUND_MUSIC_60_SMD = 841,
            SOUND_MUSIC_61_SMD = 842,
            SOUND_MUSIC_62_SMD = 843,
            SOUND_MUSIC_63_SMD = 844,
            SOUND_MUSIC_64_SMD = 845,
            SOUND_MUSIC_65_SMD = 846,
            SOUND_MUSIC_66_SMD = 847,
            SOUND_MUSIC_67_SMD = 848,
            SOUND_MUSIC_68_SMD = 849,
            SOUND_MUSIC_69_SMD = 850,
            SOUND_MUSIC_70_SMD = 851,
            SOUND_MUSIC_71_SMD = 852,
            SOUND_MUSIC_72_SMD = 853,
            SOUND_MUSIC_73_SMD = 854,
            SOUND_MUSIC_74_SMD = 855,
            SOUND_MUSIC_75_SMD = 856,
            SOUND_MUSIC_76_SMD = 857,
            SOUND_MUSIC_77_SMD = 858,
            SOUND_MUSIC_78_SMD = 859,
            SOUND_MUSIC_79_SMD = 860,
            SOUND_MUSIC_80_SMD = 861,
            SOUND_MUSIC_81_SMD = 862,
            SOUND_MUSIC_82_SMD = 863,
            SOUND_MUSIC_83_SMD = 864,
            SOUND_MUSIC_84_SMD = 865,
            SOUND_MUSIC_85_SMD = 866,
            SOUND_MUSIC_86_SMD = 867,
            SOUND_MUSIC_87_SMD = 868,
            SOUND_MUSIC_88_SMD = 869,
            SOUND_MUSIC_89_SMD = 870,
            SOUND_MUSIC_90_SMD = 871,
            SOUND_MUSIC_91_SMD = 872,
            SOUND_MUSIC_92_SMD = 873,
            SOUND_MUSIC_93_SMD = 874,
            SOUND_MUSIC_94_SMD = 875,
            SOUND_MUSIC_95_SMD = 876,
            SOUND_MUSIC_96_SMD = 877,
            SOUND_MUSIC_97_SMD = 878,
            SOUND_MUSIC_98_SMD = 879,
            SOUND_MUSIC_99_SMD = 880,
            OPEN_OPNTEX_BIN = 882,
            OPEN_OPNBK_BIN = 883,
            BATTLE_MaleDarkKnight_SPR = 885,
            BATTLE_FemaleDarkKnight_SPR = 886,
            BATTLE_MaleOnionKnight_SPR = 887,
            BATTLE_FemaleOnionKnight_SPR = 888,
            BATTLE_Balthier_SPR = 889,
            BATTLE_Luso_SPR = 890,
            BATTLE_DeathKnightArgath_SPR = 891,
            BATTLE_Aliste_SPR = 892,
            BATTLE_BremondtHuman_SPR = 893,
            BATTLE_BremondtDarkDragon_SPR = 894,
            BATTLE_SwordInStone_SPR = 895,
            BATTLE_ENTD5_ENT = 897,
            SAVEIMAGE_902_png = 902,
            SAVEIMAGE_903_png = 903,
            SAVEIMAGE_904_png = 904,
            SAVEIMAGE_905_png = 905,
            SAVEIMAGE_906_png = 906,
            SAVEIMAGE_907_png = 907,
            SAVEIMAGE_908_png = 908,
            SAVEIMAGE_909_png = 909,
            SAVEIMAGE_910_png = 910,
            SAVEIMAGE_911_png = 911,
            SAVEIMAGE_912_png = 912,
            SAVEIMAGE_913_png = 913,
            SAVEIMAGE_914_png = 914,
            SAVEIMAGE_915_png = 915,
            SAVEIMAGE_916_png = 916,
            SAVEIMAGE_917_png = 917,
            SAVEIMAGE_918_png = 918,
            SAVEIMAGE_919_png = 919,
            SAVEIMAGE_920_png = 920,
            SAVEIMAGE_921_png = 921,
            SAVEIMAGE_922_png = 922,
            SAVEIMAGE_923_png = 923,
            SAVEIMAGE_924_png = 924,
            SAVEIMAGE_925_png = 925,
            SAVEIMAGE_926_png = 926,
            SAVEIMAGE_927_png = 927,
            SAVEIMAGE_928_png = 928,
            SAVEIMAGE_929_png = 929,
            SAVEIMAGE_930_png = 930,
            SAVEIMAGE_931_png = 931,
            SAVEIMAGE_932_png = 932,
            SAVEIMAGE_933_png = 933,
            SAVEIMAGE_934_png = 934,
            SAVEIMAGE_935_png = 935,
            EFFECT_E001_BIN_duplicate = 940,
            EFFECT_E002_BIN_duplicate = 941,
            EFFECT_E003_BIN_duplicate = 942,
            EFFECT_E008_BIN_duplicate = 947,
            EFFECT_E009_BIN_duplicate = 948,
            EFFECT_E010_BIN_duplicate = 949,
            EFFECT_E011_BIN_duplicate = 950,
            EFFECT_E012_BIN_duplicate = 951,
            EFFECT_E013_BIN_duplicate = 952,
            EFFECT_E014_BIN_duplicate = 953,
            EFFECT_E016_BIN_duplicate = 955,
            EFFECT_E017_BIN_duplicate = 956,
            EFFECT_E018_BIN_duplicate = 957,
            EFFECT_E019_BIN_duplicate = 958,
            EFFECT_E020_BIN_duplicate = 959,
            EFFECT_E021_BIN_duplicate = 960,
            EFFECT_E022_BIN_duplicate = 961,
            EFFECT_E023_BIN_duplicate = 962,
            EFFECT_E024_BIN_duplicate = 963,
            EFFECT_E025_BIN_duplicate = 964,
            EFFECT_E026_BIN_duplicate = 965,
            EFFECT_E027_BIN_duplicate = 966,
            EFFECT_E028_BIN_duplicate = 967,
            EFFECT_E029_BIN_duplicate = 968,
            EFFECT_E030_BIN_duplicate = 969,
            EFFECT_E031_BIN_duplicate = 970,
            EFFECT_E032_BIN_duplicate = 971,
            EFFECT_E034_BIN_duplicate = 973,
            EFFECT_E036_BIN_duplicate = 975,
            EFFECT_E039_BIN_duplicate = 976,
            EFFECT_E040_BIN_duplicate = 977,
            EFFECT_E043_BIN_duplicate = 979,
            EFFECT_E044_BIN_duplicate = 980,
            EFFECT_E045_BIN_duplicate = 981,
            EFFECT_E046_BIN_duplicate = 982,
            EFFECT_E049_BIN_duplicate = 984,
            EFFECT_E050_BIN_duplicate = 985,
            EFFECT_E051_BIN_duplicate = 986,
            EFFECT_E052_BIN_duplicate = 987,
            EFFECT_E053_BIN_duplicate = 988,
            EFFECT_E054_BIN_duplicate = 989,
            EFFECT_E055_BIN_duplicate = 990,
            EFFECT_E056_BIN_duplicate = 991,
            EFFECT_E057_BIN_duplicate = 992,
            EFFECT_E058_BIN_duplicate = 993,
            EFFECT_E059_BIN_duplicate = 994,
            EFFECT_E060_BIN_duplicate = 995,
            EFFECT_E061_BIN_duplicate = 996,
            EFFECT_E062_BIN_duplicate = 997,
            EFFECT_E063_BIN_duplicate = 998,
            EFFECT_E066_BIN_duplicate = 1000,
            EFFECT_E078_BIN_duplicate = 1011,
            EFFECT_E081_BIN_duplicate = 1014,
            EFFECT_E082_BIN_duplicate = 1015,
            EFFECT_E083_BIN_duplicate = 1016,
            EFFECT_E084_BIN_duplicate = 1017,
            EFFECT_E085_BIN_duplicate = 1018,
            EFFECT_E086_BIN_duplicate = 1019,
            EFFECT_E087_BIN_duplicate = 1020,
            EFFECT_E095_BIN_duplicate = 1028,
            EFFECT_E096_BIN_duplicate = 1029,
            EFFECT_E097_BIN_duplicate = 1030,
            EFFECT_E098_BIN_duplicate = 1031,
            EFFECT_E099_BIN_duplicate = 1032,
            EFFECT_E100_BIN_duplicate = 1033,
            EFFECT_E101_BIN_duplicate = 1034,
            EFFECT_E102_BIN_duplicate = 1035,
            EFFECT_E103_BIN_duplicate = 1036,
            EFFECT_E104_BIN_duplicate = 1037,
            EFFECT_E105_BIN_duplicate = 1038,
            EFFECT_E106_BIN_duplicate = 1039,
            EFFECT_E107_BIN_duplicate = 1040,
            EFFECT_E108_BIN_duplicate = 1041,
            EFFECT_E109_BIN_duplicate = 1042,
            EFFECT_E110_BIN_duplicate = 1043,
            EFFECT_E111_BIN_duplicate = 1044,
            EFFECT_E112_BIN_duplicate = 1045,
            EFFECT_E113_BIN_duplicate = 1046,
            EFFECT_E114_BIN_duplicate = 1047,
            EFFECT_E115_BIN_duplicate = 1048,
            EFFECT_E116_BIN_duplicate = 1049,
            EFFECT_E117_BIN_duplicate = 1050,
            EFFECT_E118_BIN_duplicate = 1051,
            EFFECT_E119_BIN_duplicate = 1052,
            EFFECT_E120_BIN_duplicate = 1053,
            EFFECT_E121_BIN_duplicate = 1054,
            EFFECT_E122_BIN_duplicate = 1055,
            EFFECT_E123_BIN_duplicate = 1056,
            EFFECT_E124_BIN_duplicate = 1057,
            EFFECT_E125_BIN_duplicate = 1058,
            EFFECT_E126_BIN_duplicate = 1059,
            EFFECT_E127_BIN_duplicate = 1060,
            EFFECT_E128_BIN_duplicate = 1061,
            EFFECT_E129_BIN_duplicate = 1062,
            EFFECT_E130_BIN_duplicate = 1063,
            EFFECT_E131_BIN_duplicate = 1064,
            EFFECT_E132_BIN_duplicate = 1065,
            EFFECT_E133_BIN_duplicate = 1066,
            EFFECT_E135_BIN_duplicate = 1068,
            EFFECT_E136_BIN_duplicate = 1069,
            EFFECT_E137_BIN_duplicate = 1070,
            EFFECT_E139_BIN_duplicate = 1072,
            EFFECT_E140_BIN_duplicate = 1073,
            EFFECT_E141_BIN_duplicate = 1074,
            EFFECT_E142_BIN_duplicate = 1075,
            EFFECT_E154_BIN_duplicate = 1076,
            EFFECT_E156_BIN_duplicate = 1077,
            EFFECT_E157_BIN_duplicate = 1078,
            EFFECT_E158_BIN_duplicate = 1079,
            EFFECT_E159_BIN_duplicate = 1080,
            EFFECT_E160_BIN_duplicate = 1081,
            EFFECT_E163_BIN_duplicate = 1083,
            EFFECT_E164_BIN_duplicate = 1084,
            EFFECT_E165_BIN_duplicate = 1085,
            EFFECT_E166_BIN_duplicate = 1086,
            EFFECT_E168_BIN_duplicate = 1088,
            EFFECT_E169_BIN_duplicate = 1089,
            EFFECT_E170_BIN_duplicate = 1090,
            EFFECT_E171_BIN_duplicate = 1091,
            EFFECT_E172_BIN_duplicate = 1092,
            EFFECT_E173_BIN_duplicate = 1093,
            EFFECT_E175_BIN_duplicate = 1095,
            EFFECT_E177_BIN_duplicate = 1097,
            EFFECT_E178_BIN_duplicate = 1098,
            EFFECT_E179_BIN_duplicate = 1099,
            EFFECT_E180_BIN_duplicate = 1100,
            EFFECT_E181_BIN_duplicate = 1101,
            EFFECT_E182_BIN_duplicate = 1102,
            EFFECT_E183_BIN_duplicate = 1103,
            EFFECT_E184_BIN_duplicate = 1104,
            EFFECT_E185_BIN_duplicate = 1105,
            EFFECT_E186_BIN_duplicate = 1106,
            EFFECT_E187_BIN_duplicate = 1107,
            EFFECT_E188_BIN_duplicate = 1108,
            EFFECT_E189_BIN_duplicate = 1109,
            EFFECT_E190_BIN_duplicate = 1110,
            EFFECT_E191_BIN_duplicate = 1111,
            EFFECT_E192_BIN_duplicate = 1112,
            EFFECT_E194_BIN_duplicate = 1114,
            EFFECT_E195_BIN_duplicate = 1115,
            EFFECT_E196_BIN_duplicate = 1116,
            EFFECT_E197_BIN_duplicate = 1117,
            EFFECT_E198_BIN_duplicate = 1118,
            EFFECT_E199_BIN_duplicate = 1119,
            EFFECT_E200_BIN_duplicate = 1120,
            EFFECT_E201_BIN_duplicate = 1121,
            EFFECT_E202_BIN_duplicate = 1122,
            EFFECT_E203_BIN_duplicate = 1123,
            EFFECT_E204_BIN_duplicate = 1124,
            EFFECT_E205_BIN_duplicate = 1125,
            EFFECT_E206_BIN_duplicate = 1126,
            EFFECT_E208_BIN_duplicate = 1127,
            EFFECT_E210_BIN_duplicate = 1128,
            EFFECT_E211_BIN_duplicate = 1129,
            EFFECT_E212_BIN_duplicate = 1130,
            EFFECT_E213_BIN_duplicate = 1131,
            EFFECT_E214_BIN_duplicate = 1132,
            EFFECT_E215_BIN_duplicate = 1133,
            EFFECT_E216_BIN_duplicate = 1134,
            EFFECT_E217_BIN_duplicate = 1135,
            EFFECT_E218_BIN_duplicate = 1136,
            EFFECT_E219_BIN_duplicate = 1137,
            EFFECT_E220_BIN_duplicate = 1138,
            EFFECT_E221_BIN_duplicate = 1139,
            EFFECT_E223_BIN_duplicate = 1141,
            EFFECT_E224_BIN_duplicate = 1142,
            EFFECT_E225_BIN_duplicate = 1143,
            EFFECT_E226_BIN_duplicate = 1144,
            EFFECT_E231_BIN_duplicate = 1148,
            EFFECT_E232_BIN_duplicate = 1149,
            EFFECT_E234_BIN_duplicate = 1151,
            EFFECT_E237_BIN_duplicate = 1154,
            EFFECT_E238_BIN_duplicate = 1155,
            EFFECT_E239_BIN_duplicate = 1156,
            EFFECT_E243_BIN_duplicate = 1160,
            EFFECT_E260_BIN_duplicate = 1172,
            EFFECT_E261_BIN_duplicate = 1173,
            EFFECT_E262_BIN_duplicate = 1174,
            EFFECT_E263_BIN_duplicate = 1175,
            EFFECT_E264_BIN_duplicate = 1176,
            EFFECT_E265_BIN_duplicate = 1177,
            EFFECT_E266_BIN_duplicate = 1178,
            EFFECT_E267_BIN_duplicate = 1179,
            EFFECT_E268_BIN_duplicate = 1180,
            EFFECT_E269_BIN_duplicate = 1181,
            EFFECT_E270_BIN_duplicate = 1182,
            EFFECT_E271_BIN_duplicate = 1183,
            EFFECT_E272_BIN_duplicate = 1184,
            EFFECT_E273_BIN_duplicate = 1185,
            EFFECT_E288_BIN_duplicate = 1186,
            EFFECT_E289_BIN_duplicate = 1187,
            EFFECT_E290_BIN_duplicate = 1188,
            EFFECT_E291_BIN_duplicate = 1189,
            EFFECT_E292_BIN_duplicate = 1190,
            EFFECT_E293_BIN_duplicate = 1191,
            EFFECT_E294_BIN_duplicate = 1192,
            EFFECT_E295_BIN_duplicate = 1193,
            EFFECT_E306_BIN_duplicate = 1194,
            EFFECT_E307_BIN_duplicate = 1195,
            EFFECT_E308_BIN_duplicate = 1196,
            EFFECT_E309_BIN_duplicate = 1197,
            EFFECT_E310_BIN_duplicate = 1198,
            EFFECT_E311_BIN_duplicate = 1199,
            EFFECT_E312_BIN_duplicate = 1200,
            EFFECT_E313_BIN_duplicate = 1201,
            EFFECT_E316_BIN_duplicate = 1203,
            EFFECT_E318_BIN_duplicate = 1205,
            EFFECT_E319_BIN_duplicate = 1206,
            EFFECT_E320_BIN_duplicate = 1207,
            EFFECT_E321_BIN_duplicate = 1208,
            EFFECT_E322_BIN_duplicate = 1209,
            EFFECT_E323_BIN_duplicate = 1210,
            EFFECT_E324_BIN_duplicate = 1211,
            EFFECT_E325_BIN_duplicate = 1212,
            EFFECT_E326_BIN_duplicate = 1213,
            EFFECT_E327_BIN_duplicate = 1214,
            EFFECT_E328_BIN_duplicate = 1215,
            EFFECT_E329_BIN_duplicate = 1216,
            EFFECT_E330_BIN_duplicate = 1217,
            EFFECT_E331_BIN_duplicate = 1218,
            EFFECT_E332_BIN_duplicate = 1219,
            EFFECT_E333_BIN_duplicate = 1220,
            EFFECT_E334_BIN_duplicate = 1221,
            EFFECT_E335_BIN_duplicate = 1222,
            EFFECT_E337_BIN_duplicate = 1224,
            EFFECT_E339_BIN_duplicate = 1226,
            EFFECT_E340_BIN_duplicate = 1227,
            EFFECT_E342_BIN_duplicate = 1229,
            EFFECT_E343_BIN_duplicate = 1230,
            EFFECT_E344_BIN_duplicate = 1231,
            EFFECT_E345_BIN_duplicate = 1232,
            EFFECT_E346_BIN_duplicate = 1233,
            EFFECT_E347_BIN_duplicate = 1234,
            EFFECT_E348_BIN_duplicate = 1235,
            EFFECT_E349_BIN_duplicate = 1236,
            EFFECT_E350_BIN_duplicate = 1237,
            EFFECT_E351_BIN_duplicate = 1238,
            EFFECT_E352_BIN_duplicate = 1239,
            EFFECT_E353_BIN_duplicate = 1240,
            EFFECT_E354_BIN_duplicate = 1241,
            EFFECT_E356_BIN_duplicate = 1243,
            EFFECT_E357_BIN_duplicate = 1244,
            EFFECT_E358_BIN_duplicate = 1245,
            EFFECT_E359_BIN_duplicate = 1246,
            EFFECT_E360_BIN_duplicate = 1247,
            EFFECT_E361_BIN_duplicate = 1248,
            EFFECT_E362_BIN_duplicate = 1249,
            EFFECT_E363_BIN_duplicate = 1250,
            EFFECT_E364_BIN_duplicate = 1251,
            EFFECT_E365_BIN_duplicate = 1252,
            EFFECT_E366_BIN_duplicate = 1253,
            EFFECT_E367_BIN_duplicate = 1254,
            EFFECT_E368_BIN_duplicate = 1255,
            EFFECT_E369_BIN_duplicate = 1256,
            EFFECT_E370_BIN_duplicate = 1257,
            EFFECT_E372_BIN_duplicate = 1259,
            EFFECT_E374_BIN_duplicate = 1261,
            EFFECT_E375_BIN_duplicate = 1262,
            EFFECT_E377_BIN_duplicate = 1264,
            EFFECT_E378_BIN_duplicate = 1265,
            EFFECT_E379_BIN_duplicate = 1266,
            EFFECT_E380_BIN_duplicate = 1267,
            EFFECT_E386_BIN_duplicate = 1273,
            EFFECT_E387_BIN_duplicate = 1274,
            EFFECT_E388_BIN_duplicate = 1275,
            EFFECT_E389_BIN_duplicate = 1276,
            EFFECT_E390_BIN_duplicate = 1277,
            EFFECT_E391_BIN_duplicate = 1278,
            EFFECT_E392_BIN_duplicate = 1279,
            EFFECT_E393_BIN_duplicate = 1280,
            EFFECT_E394_BIN_duplicate = 1281,
            EFFECT_E396_BIN_duplicate = 1283,
            EFFECT_E397_BIN_duplicate = 1284,
            EFFECT_E398_BIN_duplicate = 1285,
            EFFECT_E399_BIN_duplicate = 1286,
            EFFECT_E400_BIN_duplicate = 1287,
            EFFECT_E404_BIN_duplicate = 1291,
            EFFECT_E405_BIN_duplicate = 1292,
            EFFECT_E406_BIN_duplicate = 1293,
            EFFECT_E407_BIN_duplicate = 1294,
            EFFECT_E410_BIN_duplicate = 1297,
            EFFECT_E413_BIN_duplicate = 1300,
            EFFECT_E457_BIN_duplicate = 1308,
            EFFECT_E459_BIN_duplicate = 1310,
            EFFECT_E470_BIN_duplicate = 1321,
            EFFECT_E479_BIN_duplicate = 1330,
            EFFECT_E484_BIN_duplicate = 1335,
            EFFECT_E486_BIN_duplicate = 1337,
            EFFECT_E509_BIN_duplicate = 1338,
            EFFECT_E510_BIN_duplicate = 1339,
            MAP_MAP000_5 = 2142,
            MAP_MAP000_GNS = 2143,
            MAP_MAP001_8 = 2144,
            MAP_MAP001_9 = 2145,
            MAP_MAP001_11 = 2146,
            MAP_MAP001_14 = 2147,
            MAP_MAP001_15 = 2148,
            MAP_MAP001_26 = 2149,
            MAP_MAP001_19 = 2150,
            MAP_MAP001_42 = 2151,
            MAP_MAP001_23 = 2152,
            MAP_MAP001_38 = 2153,
            MAP_MAP001_27 = 2154,
            MAP_MAP001_34 = 2155,
            MAP_MAP001_31 = 2156,
            MAP_MAP001_30 = 2157,
            MAP_MAP001_35 = 2158,
            MAP_MAP001_46 = 2159,
            MAP_MAP001_39 = 2160,
            MAP_MAP001_62 = 2161,
            MAP_MAP001_43 = 2162,
            MAP_MAP001_57 = 2163,
            MAP_MAP001_47 = 2164,
            MAP_MAP001_48 = 2165,
            MAP_MAP001_52 = 2166,
            MAP_MAP001_53 = 2167,
            MAP_MAP001_67 = 2168,
            MAP_MAP001_58 = 2169,
            MAP_MAP001_87 = 2170,
            MAP_MAP001_63 = 2171,
            MAP_MAP001_92 = 2172,
            MAP_MAP001_68 = 2173,
            MAP_MAP001_77 = 2174,
            MAP_MAP001_73 = 2175,
            MAP_MAP001_82 = 2176,
            MAP_MAP001_78 = 2177,
            MAP_MAP001_72 = 2178,
            MAP_MAP001_83 = 2179,
            MAP_MAP001_18 = 2180,
            MAP_MAP001_88 = 2181,
            MAP_MAP001_22 = 2182,
            MAP_MAP001_93 = 2183,
            MAP_MAP002_8 = 2185,
            MAP_MAP002_9 = 2186,
            MAP_MAP002_58 = 2187,
            MAP_MAP002_13 = 2188,
            MAP_MAP002_16 = 2189,
            MAP_MAP002_63 = 2190,
            MAP_MAP002_44 = 2191,
            MAP_MAP002_67 = 2192,
            MAP_MAP002_40 = 2193,
            MAP_MAP002_71 = 2194,
            MAP_MAP002_62 = 2195,
            MAP_MAP002_75 = 2196,
            MAP_MAP002_66 = 2197,
            MAP_MAP002_79 = 2198,
            MAP_MAP002_70 = 2199,
            MAP_MAP002_37 = 2200,
            MAP_MAP002_82 = 2201,
            MAP_MAP002_41 = 2202,
            MAP_MAP002_78 = 2203,
            MAP_MAP002_45 = 2204,
            MAP_MAP002_54 = 2205,
            MAP_MAP002_55 = 2206,
            MAP_MAP002_12 = 2207,
            MAP_MAP002_59 = 2208,
            MAP_MAP002_74 = 2209,
            MAP_MAP002_17 = 2210,
            MAP_MAP002_20 = 2211,
            MAP_MAP002_21 = 2212,
            MAP_MAP002_24 = 2213,
            MAP_MAP002_25 = 2214,
            MAP_MAP002_28 = 2215,
            MAP_MAP002_29 = 2216,
            MAP_MAP002_32 = 2217,
            MAP_MAP002_33 = 2218,
            MAP_MAP002_36 = 2219,
            MAP_MAP002_83 = 2220,
            MAP_MAP002_86 = 2221,
            MAP_MAP002_87 = 2222,
            MAP_MAP002_90 = 2223,
            MAP_MAP002_91 = 2224,
            MAP_MAP003_8 = 2227,
            MAP_MAP003_9 = 2228,
            MAP_MAP003_12 = 2229,
            MAP_MAP003_13 = 2230,
            MAP_MAP003_16 = 2231,
            MAP_MAP003_17 = 2232,
            MAP_MAP003_20 = 2233,
            MAP_MAP003_21 = 2234,
            MAP_MAP004_7 = 2237,
            MAP_MAP004_8 = 2238,
            MAP_MAP004_13 = 2239,
            MAP_MAP004_18 = 2240,
            MAP_MAP004_41 = 2241,
            MAP_MAP004_24 = 2242,
            MAP_MAP004_47 = 2243,
            MAP_MAP004_30 = 2244,
            MAP_MAP004_53 = 2245,
            MAP_MAP004_36 = 2246,
            MAP_MAP004_23 = 2247,
            MAP_MAP004_42 = 2248,
            MAP_MAP004_29 = 2249,
            MAP_MAP004_48 = 2250,
            MAP_MAP004_35 = 2251,
            MAP_MAP004_54 = 2252,
            MAP_MAP005_7 = 2255,
            MAP_MAP005_8 = 2256,
            MAP_MAP005_10 = 2257,
            MAP_MAP005_11 = 2258,
            MAP_MAP005_13 = 2259,
            MAP_MAP005_14 = 2260,
            MAP_MAP005_16 = 2261,
            MAP_MAP005_17 = 2262,
            MAP_MAP005_19 = 2263,
            MAP_MAP005_20 = 2264,
            MAP_MAP005_23 = 2265,
            MAP_MAP005_24 = 2266,
            MAP_MAP005_27 = 2267,
            MAP_MAP005_28 = 2268,
            MAP_MAP005_31 = 2269,
            MAP_MAP005_32 = 2270,
            MAP_MAP005_39 = 2271,
            MAP_MAP005_36 = 2272,
            MAP_MAP005_35 = 2273,
            MAP_MAP005_40 = 2274,
            MAP_MAP006_11 = 2277,
            MAP_MAP006_12 = 2278,
            MAP_MAP006_13 = 2279,
            MAP_MAP006_20 = 2280,
            MAP_MAP006_16 = 2281,
            MAP_MAP006_15 = 2282,
            MAP_MAP006_21 = 2283,
            MAP_MAP006_35 = 2284,
            MAP_MAP006_26 = 2285,
            MAP_MAP006_31 = 2286,
            MAP_MAP006_25 = 2287,
            MAP_MAP006_36 = 2288,
            MAP_MAP006_42 = 2289,
            MAP_MAP006_43 = 2290,
            MAP_MAP006_49 = 2291,
            MAP_MAP006_50 = 2292,
            MAP_MAP007_8 = 2295,
            MAP_MAP007_9 = 2296,
            MAP_MAP007_11 = 2297,
            MAP_MAP007_12 = 2298,
            MAP_MAP007_15 = 2299,
            MAP_MAP007_16 = 2300,
            MAP_MAP007_19 = 2301,
            MAP_MAP007_20 = 2302,
            MAP_MAP008_27 = 2305,
            MAP_MAP008_8 = 2306,
            MAP_MAP008_11 = 2307,
            MAP_MAP008_12 = 2308,
            MAP_MAP008_15 = 2309,
            MAP_MAP008_16 = 2310,
            MAP_MAP008_19 = 2311,
            MAP_MAP008_40 = 2312,
            MAP_MAP008_7 = 2313,
            MAP_MAP008_28 = 2314,
            MAP_MAP008_31 = 2315,
            MAP_MAP008_32 = 2316,
            MAP_MAP008_35 = 2317,
            MAP_MAP008_36 = 2318,
            MAP_MAP008_39 = 2319,
            MAP_MAP008_20 = 2320,
            MAP_MAP009_8 = 2323,
            MAP_MAP009_9 = 2324,
            MAP_MAP009_13 = 2325,
            MAP_MAP009_14 = 2326,
            MAP_MAP009_23 = 2327,
            MAP_MAP009_19 = 2328,
            MAP_MAP009_18 = 2329,
            MAP_MAP009_24 = 2330,
            MAP_MAP009_52 = 2331,
            MAP_MAP009_29 = 2332,
            MAP_MAP009_34 = 2333,
            MAP_MAP009_35 = 2334,
            MAP_MAP009_46 = 2335,
            MAP_MAP009_41 = 2336,
            MAP_MAP009_40 = 2337,
            MAP_MAP009_47 = 2338,
            MAP_MAP009_28 = 2339,
            MAP_MAP009_53 = 2340,
            MAP_MAP009_58 = 2341,
            MAP_MAP009_59 = 2342,
            MAP_MAP010_8 = 2344,
            MAP_MAP010_9 = 2345,
            MAP_MAP011_7 = 2347,
            MAP_MAP011_8 = 2348,
            MAP_MAP011_9 = 2349,
            MAP_MAP011_17 = 2350,
            MAP_MAP011_12 = 2351,
            MAP_MAP011_14 = 2352,
            MAP_MAP011_27 = 2353,
            MAP_MAP011_18 = 2354,
            MAP_MAP011_22 = 2355,
            MAP_MAP011_39 = 2356,
            MAP_MAP011_28 = 2357,
            MAP_MAP011_33 = 2358,
            MAP_MAP011_11 = 2359,
            MAP_MAP011_40 = 2360,
            MAP_MAP011_43 = 2361,
            MAP_MAP011_47 = 2362,
            MAP_MAP011_49 = 2363,
            MAP_MAP011_53 = 2364,
            MAP_MAP012_8 = 2367,
            MAP_MAP012_9 = 2368,
            MAP_MAP012_10 = 2369,
            MAP_MAP012_14 = 2370,
            MAP_MAP012_15 = 2371,
            MAP_MAP012_19 = 2372,
            MAP_MAP012_20 = 2373,
            MAP_MAP012_24 = 2374,
            MAP_MAP012_25 = 2375,
            MAP_MAP012_26 = 2376,
            MAP_MAP012_30 = 2377,
            MAP_MAP012_31 = 2378,
            MAP_MAP012_35 = 2379,
            MAP_MAP012_36 = 2380,
            MAP_MAP012_40 = 2381,
            MAP_MAP012_41 = 2382,
            MAP_MAP013_8 = 2385,
            MAP_MAP013_9 = 2386,
            MAP_MAP013_11 = 2387,
            MAP_MAP013_12 = 2388,
            MAP_MAP013_15 = 2389,
            MAP_MAP013_16 = 2390,
            MAP_MAP013_19 = 2391,
            MAP_MAP013_20 = 2392,
            MAP_MAP014_7 = 2395,
            MAP_MAP014_8 = 2396,
            MAP_MAP014_9 = 2397,
            MAP_MAP014_30 = 2398,
            MAP_MAP014_13 = 2399,
            MAP_MAP014_16 = 2400,
            MAP_MAP014_17 = 2401,
            MAP_MAP014_40 = 2402,
            MAP_MAP014_21 = 2403,
            MAP_MAP014_25 = 2404,
            MAP_MAP014_26 = 2405,
            MAP_MAP014_12 = 2406,
            MAP_MAP014_31 = 2407,
            MAP_MAP014_35 = 2408,
            MAP_MAP014_36 = 2409,
            MAP_MAP014_20 = 2410,
            MAP_MAP014_41 = 2411,
            MAP_MAP015_8 = 2414,
            MAP_MAP015_9 = 2415,
            MAP_MAP015_60 = 2416,
            MAP_MAP015_49 = 2417,
            MAP_MAP015_51 = 2418,
            MAP_MAP015_52 = 2419,
            MAP_MAP015_72 = 2420,
            MAP_MAP015_55 = 2421,
            MAP_MAP015_69 = 2422,
            MAP_MAP015_58 = 2423,
            MAP_MAP015_48 = 2424,
            MAP_MAP015_61 = 2425,
            MAP_MAP015_66 = 2426,
            MAP_MAP015_64 = 2427,
            MAP_MAP015_63 = 2428,
            MAP_MAP015_67 = 2429,
            MAP_MAP015_54 = 2430,
            MAP_MAP015_70 = 2431,
            MAP_MAP015_57 = 2432,
            MAP_MAP015_73 = 2433,
            MAP_MAP015_45 = 2434,
            MAP_MAP015_46 = 2435,
            MAP_MAP015_23 = 2436,
            MAP_MAP015_12 = 2437,
            MAP_MAP015_14 = 2438,
            MAP_MAP015_15 = 2439,
            MAP_MAP015_29 = 2440,
            MAP_MAP015_18 = 2441,
            MAP_MAP015_32 = 2442,
            MAP_MAP015_21 = 2443,
            MAP_MAP015_11 = 2444,
            MAP_MAP015_24 = 2445,
            MAP_MAP015_35 = 2446,
            MAP_MAP015_27 = 2447,
            MAP_MAP015_26 = 2448,
            MAP_MAP015_30 = 2449,
            MAP_MAP015_20 = 2450,
            MAP_MAP015_33 = 2451,
            MAP_MAP015_17 = 2452,
            MAP_MAP015_36 = 2453,
            MAP_MAP016_30 = 2455,
            MAP_MAP016_9 = 2456,
            MAP_MAP016_12 = 2457,
            MAP_MAP016_13 = 2458,
            MAP_MAP016_16 = 2459,
            MAP_MAP016_17 = 2460,
            MAP_MAP016_20 = 2461,
            MAP_MAP016_43 = 2462,
            MAP_MAP016_8 = 2463,
            MAP_MAP016_31 = 2464,
            MAP_MAP016_34 = 2465,
            MAP_MAP016_35 = 2466,
            MAP_MAP016_38 = 2467,
            MAP_MAP016_39 = 2468,
            MAP_MAP016_42 = 2469,
            MAP_MAP016_21 = 2470,
            MAP_MAP017_9 = 2472,
            MAP_MAP017_10 = 2473,
            MAP_MAP018_8 = 2475,
            MAP_MAP018_9 = 2476,
            MAP_MAP018_43 = 2477,
            MAP_MAP018_14 = 2478,
            MAP_MAP018_48 = 2479,
            MAP_MAP018_19 = 2480,
            MAP_MAP018_38 = 2481,
            MAP_MAP018_24 = 2482,
            MAP_MAP018_33 = 2483,
            MAP_MAP018_34 = 2484,
            MAP_MAP018_18 = 2485,
            MAP_MAP018_39 = 2486,
            MAP_MAP018_23 = 2487,
            MAP_MAP018_44 = 2488,
            MAP_MAP018_13 = 2489,
            MAP_MAP018_49 = 2490,
            MAP_MAP019_7 = 2492,
            MAP_MAP019_8 = 2493,
            MAP_MAP019_11 = 2494,
            MAP_MAP019_12 = 2495,
            MAP_MAP019_15 = 2496,
            MAP_MAP019_16 = 2497,
            MAP_MAP019_19 = 2498,
            MAP_MAP019_20 = 2499,
            MAP_MAP019_39 = 2500,
            MAP_MAP019_24 = 2501,
            MAP_MAP019_27 = 2502,
            MAP_MAP019_28 = 2503,
            MAP_MAP019_31 = 2504,
            MAP_MAP019_32 = 2505,
            MAP_MAP019_35 = 2506,
            MAP_MAP019_36 = 2507,
            MAP_MAP019_23 = 2508,
            MAP_MAP019_40 = 2509,
            MAP_MAP019_43 = 2510,
            MAP_MAP019_44 = 2511,
            MAP_MAP020_7 = 2513,
            MAP_MAP020_8 = 2514,
            MAP_MAP020_11 = 2515,
            MAP_MAP020_12 = 2516,
            MAP_MAP020_15 = 2517,
            MAP_MAP020_16 = 2518,
            MAP_MAP020_19 = 2519,
            MAP_MAP020_20 = 2520,
            MAP_MAP021_44 = 2523,
            MAP_MAP021_8 = 2524,
            MAP_MAP021_15 = 2525,
            MAP_MAP021_18 = 2526,
            MAP_MAP021_19 = 2527,
            MAP_MAP021_22 = 2528,
            MAP_MAP021_23 = 2529,
            MAP_MAP021_50 = 2530,
            MAP_MAP021_27 = 2531,
            MAP_MAP021_33 = 2532,
            MAP_MAP021_38 = 2533,
            MAP_MAP021_39 = 2534,
            MAP_MAP021_7 = 2535,
            MAP_MAP021_45 = 2536,
            MAP_MAP021_26 = 2537,
            MAP_MAP021_51 = 2538,
            MAP_MAP022_8 = 2541,
            MAP_MAP022_9 = 2542,
            MAP_MAP022_12 = 2543,
            MAP_MAP022_13 = 2544,
            MAP_MAP022_16 = 2545,
            MAP_MAP022_17 = 2546,
            MAP_MAP022_20 = 2547,
            MAP_MAP022_21 = 2548,
            MAP_MAP022_24 = 2549,
            MAP_MAP022_25 = 2550,
            MAP_MAP022_30 = 2551,
            MAP_MAP022_31 = 2552,
            MAP_MAP022_36 = 2553,
            MAP_MAP022_37 = 2554,
            MAP_MAP022_48 = 2555,
            MAP_MAP022_43 = 2556,
            MAP_MAP022_42 = 2557,
            MAP_MAP022_49 = 2558,
            MAP_MAP022_54 = 2559,
            MAP_MAP022_55 = 2560,
            MAP_MAP023_8 = 2563,
            MAP_MAP023_9 = 2564,
            MAP_MAP023_10 = 2565,
            MAP_MAP023_12 = 2566,
            MAP_MAP023_13 = 2567,
            MAP_MAP023_15 = 2568,
            MAP_MAP023_16 = 2569,
            MAP_MAP023_19 = 2570,
            MAP_MAP023_20 = 2571,
            MAP_MAP023_24 = 2572,
            MAP_MAP023_25 = 2573,
            MAP_MAP024_6 = 2576,
            MAP_MAP024_7 = 2577,
            MAP_MAP024_10 = 2578,
            MAP_MAP024_11 = 2579,
            MAP_MAP024_14 = 2580,
            MAP_MAP024_15 = 2581,
            MAP_MAP024_18 = 2582,
            MAP_MAP024_19 = 2583,
            MAP_MAP025_7 = 2586,
            MAP_MAP025_8 = 2587,
            MAP_MAP025_9 = 2588,
            MAP_MAP025_12 = 2589,
            MAP_MAP025_13 = 2590,
            MAP_MAP025_20 = 2591,
            MAP_MAP025_17 = 2592,
            MAP_MAP025_16 = 2593,
            MAP_MAP025_21 = 2594,
            MAP_MAP025_24 = 2595,
            MAP_MAP025_25 = 2596,
            MAP_MAP026_8 = 2599,
            MAP_MAP026_9 = 2600,
            MAP_MAP027_7 = 2603,
            MAP_MAP027_8 = 2604,
            MAP_MAP027_10 = 2605,
            MAP_MAP027_13 = 2606,
            MAP_MAP027_14 = 2607,
            MAP_MAP027_21 = 2608,
            MAP_MAP027_18 = 2609,
            MAP_MAP027_17 = 2610,
            MAP_MAP027_22 = 2611,
            MAP_MAP028_6 = 2614,
            MAP_MAP028_7 = 2615,
            MAP_MAP028_8 = 2616,
            MAP_MAP028_10 = 2617,
            MAP_MAP028_11 = 2618,
            MAP_MAP028_13 = 2619,
            MAP_MAP028_14 = 2620,
            MAP_MAP028_16 = 2621,
            MAP_MAP028_17 = 2622,
            MAP_MAP029_7 = 2625,
            MAP_MAP029_8 = 2626,
            MAP_MAP029_10 = 2627,
            MAP_MAP029_13 = 2628,
            MAP_MAP029_14 = 2629,
            MAP_MAP029_21 = 2630,
            MAP_MAP029_18 = 2631,
            MAP_MAP029_17 = 2632,
            MAP_MAP029_22 = 2633,
            MAP_MAP030_6 = 2636,
            MAP_MAP030_7 = 2637,
            MAP_MAP030_8 = 2638,
            MAP_MAP030_11 = 2639,
            MAP_MAP030_12 = 2640,
            MAP_MAP030_19 = 2641,
            MAP_MAP030_16 = 2642,
            MAP_MAP030_15 = 2643,
            MAP_MAP030_20 = 2644,
            MAP_MAP031_7 = 2647,
            MAP_MAP031_8 = 2648,
            MAP_MAP031_11 = 2649,
            MAP_MAP031_12 = 2650,
            MAP_MAP031_15 = 2651,
            MAP_MAP031_16 = 2652,
            MAP_MAP031_19 = 2653,
            MAP_MAP031_20 = 2654,
            MAP_MAP031_23 = 2655,
            MAP_MAP031_24 = 2656,
            MAP_MAP031_27 = 2657,
            MAP_MAP031_28 = 2658,
            MAP_MAP031_43 = 2659,
            MAP_MAP031_32 = 2660,
            MAP_MAP031_31 = 2661,
            MAP_MAP031_36 = 2662,
            MAP_MAP031_35 = 2663,
            MAP_MAP031_40 = 2664,
            MAP_MAP031_39 = 2665,
            MAP_MAP031_44 = 2666,
            MAP_MAP032_8 = 2669,
            MAP_MAP032_9 = 2670,
            MAP_MAP032_11 = 2671,
            MAP_MAP032_14 = 2672,
            MAP_MAP032_15 = 2673,
            MAP_MAP032_22 = 2674,
            MAP_MAP032_19 = 2675,
            MAP_MAP032_18 = 2676,
            MAP_MAP032_23 = 2677,
            MAP_MAP033_8 = 2680,
            MAP_MAP033_9 = 2681,
            MAP_MAP033_18 = 2682,
            MAP_MAP033_19 = 2683,
            MAP_MAP034_6 = 2685,
            MAP_MAP034_7 = 2686,
            MAP_MAP034_11 = 2687,
            MAP_MAP034_15 = 2688,
            MAP_MAP034_19 = 2689,
            MAP_MAP035_7 = 2692,
            MAP_MAP035_8 = 2693,
            MAP_MAP035_11 = 2694,
            MAP_MAP035_12 = 2695,
            MAP_MAP035_15 = 2696,
            MAP_MAP035_16 = 2697,
            MAP_MAP035_19 = 2698,
            MAP_MAP035_20 = 2699,
            MAP_MAP035_23 = 2700,
            MAP_MAP035_24 = 2701,
            MAP_MAP035_27 = 2702,
            MAP_MAP035_28 = 2703,
            MAP_MAP035_39 = 2704,
            MAP_MAP035_32 = 2705,
            MAP_MAP035_35 = 2706,
            MAP_MAP035_36 = 2707,
            MAP_MAP035_31 = 2708,
            MAP_MAP035_40 = 2709,
            MAP_MAP035_43 = 2710,
            MAP_MAP035_44 = 2711,
            MAP_MAP036_1 = 2714,
            MAP_MAP036_4 = 2715,
            MAP_MAP036_5 = 2716,
            MAP_MAP036_8 = 2717,
            MAP_MAP036_9 = 2718,
            MAP_MAP036_12 = 2719,
            MAP_MAP036_13 = 2720,
            MAP_MAP036_16 = 2721,
            MAP_MAP036_17 = 2722,
            MAP_MAP036_20 = 2723,
            MAP_MAP036_21 = 2724,
            MAP_MAP036_24 = 2725,
            MAP_MAP036_25 = 2726,
            MAP_MAP036_28 = 2727,
            MAP_MAP036_29 = 2728,
            MAP_MAP036_32 = 2729,
            MAP_MAP036_33 = 2730,
            MAP_MAP036_45 = 2731,
            MAP_MAP036_46 = 2732,
            MAP_MAP037_6 = 2735,
            MAP_MAP037_7 = 2736,
            MAP_MAP037_9 = 2737,
            MAP_MAP037_10 = 2738,
            MAP_MAP038_11 = 2741,
            MAP_MAP038_12 = 2742,
            MAP_MAP038_15 = 2743,
            MAP_MAP038_16 = 2744,
            MAP_MAP038_19 = 2745,
            MAP_MAP038_20 = 2746,
            MAP_MAP038_27 = 2747,
            MAP_MAP038_24 = 2748,
            MAP_MAP038_23 = 2749,
            MAP_MAP038_28 = 2750,
            MAP_MAP038_31 = 2751,
            MAP_MAP038_32 = 2752,
            MAP_MAP038_43 = 2753,
            MAP_MAP038_36 = 2754,
            MAP_MAP038_35 = 2755,
            MAP_MAP038_40 = 2756,
            MAP_MAP038_39 = 2757,
            MAP_MAP038_44 = 2758,
            MAP_MAP038_47 = 2759,
            MAP_MAP038_48 = 2760,
            MAP_MAP039_8 = 2763,
            MAP_MAP039_9 = 2764,
            MAP_MAP040_9 = 2767,
            MAP_MAP040_10 = 2768,
            MAP_MAP040_11 = 2769,
            MAP_MAP040_19 = 2770,
            MAP_MAP040_14 = 2771,
            MAP_MAP040_16 = 2772,
            MAP_MAP040_17 = 2773,
            MAP_MAP040_13 = 2774,
            MAP_MAP040_20 = 2775,
            MAP_MAP040_34 = 2776,
            MAP_MAP040_23 = 2777,
            MAP_MAP040_31 = 2778,
            MAP_MAP040_26 = 2779,
            MAP_MAP040_28 = 2780,
            MAP_MAP040_29 = 2781,
            MAP_MAP040_25 = 2782,
            MAP_MAP040_32 = 2783,
            MAP_MAP040_22 = 2784,
            MAP_MAP040_35 = 2785,
            MAP_MAP041_5 = 2788,
            MAP_MAP041_11 = 2789,
            MAP_MAP041_19 = 2790,
            MAP_MAP041_20 = 2791,
            MAP_MAP042_8 = 2794,
            MAP_MAP042_9 = 2795,
            MAP_MAP042_12 = 2796,
            MAP_MAP042_13 = 2797,
            MAP_MAP042_16 = 2798,
            MAP_MAP042_17 = 2799,
            MAP_MAP042_20 = 2800,
            MAP_MAP042_21 = 2801,
            MAP_MAP042_24 = 2802,
            MAP_MAP042_25 = 2803,
            MAP_MAP043_8 = 2805,
            MAP_MAP043_9 = 2806,
            MAP_MAP043_12 = 2807,
            MAP_MAP043_13 = 2808,
            MAP_MAP044_7 = 2810,
            MAP_MAP044_8 = 2811,
            MAP_MAP044_15 = 2812,
            MAP_MAP044_12 = 2813,
            MAP_MAP044_11 = 2814,
            MAP_MAP044_16 = 2815,
            MAP_MAP044_23 = 2816,
            MAP_MAP044_20 = 2817,
            MAP_MAP044_19 = 2818,
            MAP_MAP044_24 = 2819,
            MAP_MAP044_27 = 2820,
            MAP_MAP044_28 = 2821,
            MAP_MAP044_31 = 2822,
            MAP_MAP044_32 = 2823,
            MAP_MAP044_39 = 2824,
            MAP_MAP044_36 = 2825,
            MAP_MAP044_35 = 2826,
            MAP_MAP044_40 = 2827,
            MAP_MAP044_43 = 2828,
            MAP_MAP044_44 = 2829,
            MAP_MAP045_6 = 2832,
            MAP_MAP045_7 = 2833,
            MAP_MAP045_10 = 2834,
            MAP_MAP045_11 = 2835,
            MAP_MAP046_8 = 2838,
            MAP_MAP046_9 = 2839,
            MAP_MAP046_12 = 2840,
            MAP_MAP046_13 = 2841,
            MAP_MAP046_16 = 2842,
            MAP_MAP046_17 = 2843,
            MAP_MAP046_20 = 2844,
            MAP_MAP046_21 = 2845,
            MAP_MAP046_24 = 2846,
            MAP_MAP046_25 = 2847,
            MAP_MAP047_16 = 2849,
            MAP_MAP047_17 = 2850,
            MAP_MAP047_20 = 2851,
            MAP_MAP047_21 = 2852,
            MAP_MAP047_25 = 2853,
            MAP_MAP047_26 = 2854,
            MAP_MAP047_35 = 2855,
            MAP_MAP047_31 = 2856,
            MAP_MAP047_30 = 2857,
            MAP_MAP047_36 = 2858,
            MAP_MAP047_40 = 2859,
            MAP_MAP047_41 = 2860,
            MAP_MAP047_60 = 2861,
            MAP_MAP047_46 = 2862,
            MAP_MAP047_55 = 2863,
            MAP_MAP047_51 = 2864,
            MAP_MAP047_50 = 2865,
            MAP_MAP047_56 = 2866,
            MAP_MAP047_45 = 2867,
            MAP_MAP047_61 = 2868,
            MAP_MAP048_8 = 2871,
            MAP_MAP048_9 = 2872,
            MAP_MAP048_12 = 2873,
            MAP_MAP048_13 = 2874,
            MAP_MAP048_16 = 2875,
            MAP_MAP048_17 = 2876,
            MAP_MAP048_24 = 2877,
            MAP_MAP048_21 = 2878,
            MAP_MAP048_20 = 2879,
            MAP_MAP048_25 = 2880,
            MAP_MAP048_28 = 2881,
            MAP_MAP048_29 = 2882,
            MAP_MAP048_32 = 2883,
            MAP_MAP048_33 = 2884,
            MAP_MAP048_36 = 2885,
            MAP_MAP048_37 = 2886,
            MAP_MAP048_40 = 2887,
            MAP_MAP048_41 = 2888,
            MAP_MAP048_44 = 2889,
            MAP_MAP048_45 = 2890,
            MAP_MAP049_7 = 2892,
            MAP_MAP049_8 = 2893,
            MAP_MAP049_10 = 2894,
            MAP_MAP049_11 = 2895,
            MAP_MAP049_14 = 2896,
            MAP_MAP049_15 = 2897,
            MAP_MAP049_22 = 2898,
            MAP_MAP049_19 = 2899,
            MAP_MAP049_18 = 2900,
            MAP_MAP049_23 = 2901,
            MAP_MAP049_26 = 2902,
            MAP_MAP049_27 = 2903,
            MAP_MAP049_30 = 2904,
            MAP_MAP049_31 = 2905,
            MAP_MAP049_34 = 2906,
            MAP_MAP049_35 = 2907,
            MAP_MAP049_38 = 2908,
            MAP_MAP049_39 = 2909,
            MAP_MAP049_42 = 2910,
            MAP_MAP049_43 = 2911,
            MAP_MAP050_7 = 2914,
            MAP_MAP050_8 = 2915,
            MAP_MAP050_11 = 2916,
            MAP_MAP050_12 = 2917,
            MAP_MAP050_15 = 2918,
            MAP_MAP050_16 = 2919,
            MAP_MAP050_19 = 2920,
            MAP_MAP050_20 = 2921,
            MAP_MAP050_23 = 2922,
            MAP_MAP050_24 = 2923,
            MAP_MAP050_27 = 2924,
            MAP_MAP050_28 = 2925,
            MAP_MAP050_31 = 2926,
            MAP_MAP050_32 = 2927,
            MAP_MAP050_35 = 2928,
            MAP_MAP050_36 = 2929,
            MAP_MAP050_39 = 2930,
            MAP_MAP050_40 = 2931,
            MAP_MAP050_43 = 2932,
            MAP_MAP050_44 = 2933,
            MAP_MAP051_7 = 2936,
            MAP_MAP051_8 = 2937,
            MAP_MAP051_16 = 2938,
            MAP_MAP051_17 = 2939,
            MAP_MAP051_37 = 2940,
            MAP_MAP051_21 = 2941,
            MAP_MAP051_43 = 2942,
            MAP_MAP051_25 = 2943,
            MAP_MAP051_28 = 2944,
            MAP_MAP051_29 = 2945,
            MAP_MAP051_32 = 2946,
            MAP_MAP051_20 = 2947,
            MAP_MAP051_38 = 2948,
            MAP_MAP051_24 = 2949,
            MAP_MAP051_44 = 2950,
            MAP_MAP051_49 = 2951,
            MAP_MAP051_50 = 2952,
            MAP_MAP052_9 = 2955,
            MAP_MAP052_10 = 2956,
            MAP_MAP053_7 = 2959,
            MAP_MAP053_8 = 2960,
            MAP_MAP053_10 = 2961,
            MAP_MAP053_19 = 2962,
            MAP_MAP053_22 = 2963,
            MAP_MAP054_7 = 2966,
            MAP_MAP054_8 = 2967,
            MAP_MAP055_12 = 2970,
            MAP_MAP055_13 = 2971,
            MAP_MAP055_15 = 2972,
            MAP_MAP055_14 = 2973,
            MAP_MAP056_0 = 2976,
            MAP_MAP056_3 = 2977,
            MAP_MAP056_4 = 2978,
            MAP_MAP056_7 = 2979,
            MAP_MAP056_8 = 2980,
            MAP_MAP056_11 = 2981,
            MAP_MAP056_12 = 2982,
            MAP_MAP056_16 = 2983,
            MAP_MAP056_17 = 2984,
            MAP_MAP056_26 = 2985,
            MAP_MAP056_22 = 2986,
            MAP_MAP056_31 = 2987,
            MAP_MAP056_27 = 2988,
            MAP_MAP056_21 = 2989,
            MAP_MAP056_32 = 2990,
            MAP_MAP056_36 = 2991,
            MAP_MAP056_37 = 2992,
            MAP_MAP056_47 = 2993,
            MAP_MAP056_48 = 2994,
            MAP_MAP057_8 = 2997,
            MAP_MAP057_9 = 2998,
            MAP_MAP058_8 = 3000,
            MAP_MAP058_9 = 3001,
            MAP_MAP059_8 = 3003,
            MAP_MAP059_9 = 3004,
            MAP_MAP060_8 = 3006,
            MAP_MAP060_9 = 3007,
            MAP_MAP061_7 = 3010,
            MAP_MAP061_8 = 3011,
            MAP_MAP061_9 = 3012,
            MAP_MAP061_10 = 3013,
            MAP_MAP062_7 = 3015,
            MAP_MAP062_11 = 3017,
            MAP_MAP062_12 = 3018,
            MAP_MAP062_15 = 3019,
            MAP_MAP062_16 = 3020,
            MAP_MAP062_19 = 3021,
            MAP_MAP062_20 = 3022,
            MAP_MAP063_6 = 3025,
            MAP_MAP063_7 = 3026,
            MAP_MAP063_10 = 3027,
            MAP_MAP063_11 = 3028,
            MAP_MAP063_14 = 3029,
            MAP_MAP063_15 = 3030,
            MAP_MAP063_22 = 3031,
            MAP_MAP063_19 = 3032,
            MAP_MAP063_18 = 3033,
            MAP_MAP063_23 = 3034,
            MAP_MAP064_9 = 3036,
            MAP_MAP064_10 = 3037,
            MAP_MAP064_14 = 3038,
            MAP_MAP064_19 = 3039,
            MAP_MAP065_8 = 3042,
            MAP_MAP065_9 = 3043,
            MAP_MAP066_7 = 3045,
            MAP_MAP066_8 = 3046,
            MAP_MAP066_9 = 3047,
            MAP_MAP066_12 = 3048,
            MAP_MAP066_13 = 3049,
            MAP_MAP066_32 = 3050,
            MAP_MAP066_17 = 3051,
            MAP_MAP066_40 = 3052,
            MAP_MAP066_21 = 3053,
            MAP_MAP066_24 = 3054,
            MAP_MAP066_25 = 3055,
            MAP_MAP066_36 = 3056,
            MAP_MAP066_29 = 3057,
            MAP_MAP066_16 = 3058,
            MAP_MAP066_33 = 3059,
            MAP_MAP066_20 = 3060,
            MAP_MAP066_37 = 3061,
            MAP_MAP066_28 = 3062,
            MAP_MAP066_41 = 3063,
            MAP_MAP067_7 = 3066,
            MAP_MAP067_8 = 3067,
            MAP_MAP067_9 = 3068,
            MAP_MAP067_12 = 3069,
            MAP_MAP067_13 = 3070,
            MAP_MAP067_40 = 3071,
            MAP_MAP067_17 = 3072,
            MAP_MAP067_36 = 3073,
            MAP_MAP067_21 = 3074,
            MAP_MAP067_24 = 3075,
            MAP_MAP067_25 = 3076,
            MAP_MAP067_32 = 3077,
            MAP_MAP067_29 = 3078,
            MAP_MAP067_28 = 3079,
            MAP_MAP067_33 = 3080,
            MAP_MAP067_20 = 3081,
            MAP_MAP067_37 = 3082,
            MAP_MAP067_16 = 3083,
            MAP_MAP067_41 = 3084,
            MAP_MAP068_28 = 3086,
            MAP_MAP068_8 = 3087,
            MAP_MAP068_32 = 3088,
            MAP_MAP068_12 = 3089,
            MAP_MAP068_15 = 3090,
            MAP_MAP068_16 = 3091,
            MAP_MAP068_40 = 3092,
            MAP_MAP068_41 = 3093,
            MAP_MAP068_7 = 3094,
            MAP_MAP068_29 = 3095,
            MAP_MAP068_11 = 3096,
            MAP_MAP068_33 = 3097,
            MAP_MAP068_36 = 3098,
            MAP_MAP068_37 = 3099,
            MAP_MAP068_19 = 3100,
            MAP_MAP068_20 = 3101,
            MAP_MAP069_7 = 3104,
            MAP_MAP069_8 = 3105,
            MAP_MAP070_6 = 3108,
            MAP_MAP070_7 = 3109,
            MAP_MAP070_10 = 3110,
            MAP_MAP070_11 = 3111,
            MAP_MAP071_8 = 3114,
            MAP_MAP071_9 = 3115,
            MAP_MAP071_28 = 3116,
            MAP_MAP071_13 = 3117,
            MAP_MAP071_32 = 3118,
            MAP_MAP071_17 = 3119,
            MAP_MAP071_36 = 3120,
            MAP_MAP071_21 = 3121,
            MAP_MAP071_40 = 3122,
            MAP_MAP071_25 = 3123,
            MAP_MAP071_12 = 3124,
            MAP_MAP071_29 = 3125,
            MAP_MAP071_16 = 3126,
            MAP_MAP071_33 = 3127,
            MAP_MAP071_24 = 3128,
            MAP_MAP071_37 = 3129,
            MAP_MAP071_20 = 3130,
            MAP_MAP071_41 = 3131,
            MAP_MAP071_44 = 3132,
            MAP_MAP071_45 = 3133,
            MAP_MAP072_13 = 3136,
            MAP_MAP072_9 = 3137,
            MAP_MAP072_10 = 3138,
            MAP_MAP072_8 = 3139,
            MAP_MAP072_14 = 3140,
            MAP_MAP072_17 = 3141,
            MAP_MAP072_18 = 3142,
            MAP_MAP072_34 = 3143,
            MAP_MAP072_21 = 3144,
            MAP_MAP072_23 = 3145,
            MAP_MAP072_20 = 3146,
            MAP_MAP072_27 = 3147,
            MAP_MAP072_38 = 3148,
            MAP_MAP072_31 = 3149,
            MAP_MAP072_26 = 3150,
            MAP_MAP072_35 = 3151,
            MAP_MAP072_30 = 3152,
            MAP_MAP072_39 = 3153,
            MAP_MAP073_2 = 3156,
            MAP_MAP073_3 = 3157,
            MAP_MAP073_6 = 3158,
            MAP_MAP073_7 = 3159,
            MAP_MAP073_10 = 3160,
            MAP_MAP073_11 = 3161,
            MAP_MAP073_23 = 3162,
            MAP_MAP073_24 = 3163,
            MAP_MAP074_8 = 3166,
            MAP_MAP074_9 = 3167,
            MAP_MAP074_10 = 3168,
            MAP_MAP074_21 = 3169,
            MAP_MAP074_13 = 3170,
            MAP_MAP074_27 = 3171,
            MAP_MAP074_16 = 3172,
            MAP_MAP074_15 = 3173,
            MAP_MAP074_19 = 3174,
            MAP_MAP074_12 = 3175,
            MAP_MAP074_22 = 3176,
            MAP_MAP074_24 = 3177,
            MAP_MAP074_25 = 3178,
            MAP_MAP074_18 = 3179,
            MAP_MAP074_28 = 3180,
            MAP_MAP074_33 = 3181,
            MAP_MAP074_31 = 3182,
            MAP_MAP074_30 = 3183,
            MAP_MAP074_34 = 3184,
            MAP_MAP075_8 = 3187,
            MAP_MAP075_9 = 3188,
            MAP_MAP075_11 = 3189,
            MAP_MAP075_12 = 3190,
            MAP_MAP075_25 = 3191,
            MAP_MAP075_15 = 3192,
            MAP_MAP075_45 = 3193,
            MAP_MAP075_18 = 3194,
            MAP_MAP075_40 = 3195,
            MAP_MAP075_21 = 3196,
            MAP_MAP075_14 = 3197,
            MAP_MAP075_26 = 3198,
            MAP_MAP075_35 = 3199,
            MAP_MAP075_31 = 3200,
            MAP_MAP075_20 = 3201,
            MAP_MAP075_36 = 3202,
            MAP_MAP075_30 = 3203,
            MAP_MAP075_41 = 3204,
            MAP_MAP075_17 = 3205,
            MAP_MAP075_46 = 3206,
            MAP_MAP076_6 = 3209,
            MAP_MAP076_7 = 3210,
            MAP_MAP076_9 = 3211,
            MAP_MAP076_11 = 3212,
            MAP_MAP076_24 = 3213,
            MAP_MAP076_15 = 3214,
            MAP_MAP076_18 = 3215,
            MAP_MAP076_19 = 3216,
            MAP_MAP076_21 = 3217,
            MAP_MAP076_14 = 3218,
            MAP_MAP076_25 = 3219,
            MAP_MAP076_32 = 3220,
            MAP_MAP076_29 = 3221,
            MAP_MAP076_28 = 3222,
            MAP_MAP076_33 = 3223,
            MAP_MAP076_36 = 3224,
            MAP_MAP076_37 = 3225,
            MAP_MAP077_8 = 3228,
            MAP_MAP077_9 = 3229,
            MAP_MAP077_11 = 3230,
            MAP_MAP077_12 = 3231,
            MAP_MAP077_14 = 3232,
            MAP_MAP077_15 = 3233,
            MAP_MAP077_17 = 3234,
            MAP_MAP077_18 = 3235,
            MAP_MAP077_20 = 3236,
            MAP_MAP077_21 = 3237,
            MAP_MAP077_25 = 3238,
            MAP_MAP077_26 = 3239,
            MAP_MAP077_30 = 3240,
            MAP_MAP077_31 = 3241,
            MAP_MAP077_35 = 3242,
            MAP_MAP077_36 = 3243,
            MAP_MAP077_40 = 3244,
            MAP_MAP077_41 = 3245,
            MAP_MAP077_45 = 3246,
            MAP_MAP077_46 = 3247,
            MAP_MAP078_8 = 3249,
            MAP_MAP078_9 = 3250,
            MAP_MAP078_18 = 3251,
            MAP_MAP078_14 = 3252,
            MAP_MAP078_13 = 3253,
            MAP_MAP078_19 = 3254,
            MAP_MAP078_26 = 3255,
            MAP_MAP078_23 = 3256,
            MAP_MAP078_22 = 3257,
            MAP_MAP078_27 = 3258,
            MAP_MAP078_31 = 3259,
            MAP_MAP078_32 = 3260,
            MAP_MAP078_47 = 3261,
            MAP_MAP078_36 = 3262,
            MAP_MAP078_43 = 3263,
            MAP_MAP078_40 = 3264,
            MAP_MAP078_39 = 3265,
            MAP_MAP078_44 = 3266,
            MAP_MAP078_35 = 3267,
            MAP_MAP078_48 = 3268,
            MAP_MAP079_6 = 3271,
            MAP_MAP079_7 = 3272,
            MAP_MAP079_8 = 3273,
            MAP_MAP079_9 = 3274,
            MAP_MAP079_10 = 3275,
            MAP_MAP079_11 = 3276,
            MAP_MAP079_12 = 3277,
            MAP_MAP079_13 = 3278,
            MAP_MAP079_14 = 3279,
            MAP_MAP079_15 = 3280,
            MAP_MAP079_16 = 3281,
            MAP_MAP080_6 = 3283,
            MAP_MAP080_7 = 3284,
            MAP_MAP080_8 = 3285,
            MAP_MAP080_15 = 3286,
            MAP_MAP080_12 = 3287,
            MAP_MAP080_19 = 3288,
            MAP_MAP080_16 = 3289,
            MAP_MAP080_35 = 3290,
            MAP_MAP080_20 = 3291,
            MAP_MAP080_31 = 3292,
            MAP_MAP080_24 = 3293,
            MAP_MAP080_27 = 3294,
            MAP_MAP080_28 = 3295,
            MAP_MAP080_23 = 3296,
            MAP_MAP080_32 = 3297,
            MAP_MAP080_39 = 3298,
            MAP_MAP080_36 = 3299,
            MAP_MAP080_11 = 3300,
            MAP_MAP080_40 = 3301,
            MAP_MAP081_6 = 3304,
            MAP_MAP081_7 = 3305,
            MAP_MAP081_10 = 3306,
            MAP_MAP081_11 = 3307,
            MAP_MAP081_26 = 3308,
            MAP_MAP081_15 = 3309,
            MAP_MAP081_38 = 3310,
            MAP_MAP081_19 = 3311,
            MAP_MAP081_34 = 3312,
            MAP_MAP081_23 = 3313,
            MAP_MAP081_14 = 3314,
            MAP_MAP081_27 = 3315,
            MAP_MAP081_42 = 3316,
            MAP_MAP081_31 = 3317,
            MAP_MAP081_30 = 3318,
            MAP_MAP081_35 = 3319,
            MAP_MAP081_22 = 3320,
            MAP_MAP081_39 = 3321,
            MAP_MAP081_18 = 3322,
            MAP_MAP081_43 = 3323,
            MAP_MAP082_6 = 3326,
            MAP_MAP082_7 = 3327,
            MAP_MAP082_9 = 3328,
            MAP_MAP082_11 = 3329,
            MAP_MAP082_32 = 3330,
            MAP_MAP082_15 = 3331,
            MAP_MAP082_36 = 3332,
            MAP_MAP082_19 = 3333,
            MAP_MAP082_21 = 3334,
            MAP_MAP082_24 = 3335,
            MAP_MAP082_25 = 3336,
            MAP_MAP082_28 = 3337,
            MAP_MAP082_29 = 3338,
            MAP_MAP082_18 = 3339,
            MAP_MAP082_33 = 3340,
            MAP_MAP082_14 = 3341,
            MAP_MAP082_37 = 3342,
            MAP_MAP083_8 = 3345,
            MAP_MAP083_9 = 3346,
            MAP_MAP083_10 = 3347,
            MAP_MAP083_12 = 3348,
            MAP_MAP083_14 = 3349,
            MAP_MAP083_16 = 3350,
            MAP_MAP083_18 = 3351,
            MAP_MAP083_20 = 3352,
            MAP_MAP083_22 = 3353,
            MAP_MAP083_24 = 3354,
            MAP_MAP083_26 = 3355,
            MAP_MAP083_28 = 3356,
            MAP_MAP083_38 = 3357,
            MAP_MAP084_6 = 3360,
            MAP_MAP084_7 = 3361,
            MAP_MAP084_14 = 3362,
            MAP_MAP084_11 = 3363,
            MAP_MAP084_26 = 3364,
            MAP_MAP084_15 = 3365,
            MAP_MAP084_38 = 3366,
            MAP_MAP084_19 = 3367,
            MAP_MAP084_42 = 3368,
            MAP_MAP084_23 = 3369,
            MAP_MAP084_10 = 3370,
            MAP_MAP084_27 = 3371,
            MAP_MAP084_18 = 3372,
            MAP_MAP084_31 = 3373,
            MAP_MAP084_22 = 3374,
            MAP_MAP084_35 = 3375,
            MAP_MAP084_30 = 3376,
            MAP_MAP084_39 = 3377,
            MAP_MAP084_34 = 3378,
            MAP_MAP084_43 = 3379,
            MAP_MAP085_6 = 3382,
            MAP_MAP085_7 = 3383,
            MAP_MAP085_9 = 3384,
            MAP_MAP085_10 = 3385,
            MAP_MAP085_21 = 3386,
            MAP_MAP085_13 = 3387,
            MAP_MAP085_33 = 3388,
            MAP_MAP085_16 = 3389,
            MAP_MAP085_18 = 3390,
            MAP_MAP085_19 = 3391,
            MAP_MAP085_12 = 3392,
            MAP_MAP085_22 = 3393,
            MAP_MAP085_27 = 3394,
            MAP_MAP085_25 = 3395,
            MAP_MAP085_24 = 3396,
            MAP_MAP085_28 = 3397,
            MAP_MAP085_30 = 3398,
            MAP_MAP085_31 = 3399,
            MAP_MAP085_15 = 3400,
            MAP_MAP085_34 = 3401,
            MAP_MAP086_6 = 3404,
            MAP_MAP086_7 = 3405,
            MAP_MAP086_9 = 3406,
            MAP_MAP086_10 = 3407,
            MAP_MAP086_12 = 3408,
            MAP_MAP086_13 = 3409,
            MAP_MAP086_27 = 3410,
            MAP_MAP086_16 = 3411,
            MAP_MAP086_30 = 3412,
            MAP_MAP086_19 = 3413,
            MAP_MAP086_21 = 3414,
            MAP_MAP086_22 = 3415,
            MAP_MAP086_24 = 3416,
            MAP_MAP086_25 = 3417,
            MAP_MAP086_33 = 3418,
            MAP_MAP086_28 = 3419,
            MAP_MAP086_15 = 3420,
            MAP_MAP086_31 = 3421,
            MAP_MAP086_18 = 3422,
            MAP_MAP086_34 = 3423,
            MAP_MAP087_8 = 3426,
            MAP_MAP087_9 = 3427,
            MAP_MAP087_10 = 3428,
            MAP_MAP087_11 = 3429,
            MAP_MAP087_12 = 3430,
            MAP_MAP087_13 = 3431,
            MAP_MAP087_14 = 3432,
            MAP_MAP087_15 = 3433,
            MAP_MAP087_16 = 3434,
            MAP_MAP087_17 = 3435,
            MAP_MAP087_18 = 3436,
            MAP_MAP088_8 = 3439,
            MAP_MAP088_9 = 3440,
            MAP_MAP088_23 = 3441,
            MAP_MAP088_12 = 3442,
            MAP_MAP088_14 = 3443,
            MAP_MAP088_15 = 3444,
            MAP_MAP088_20 = 3445,
            MAP_MAP088_18 = 3446,
            MAP_MAP088_17 = 3447,
            MAP_MAP088_21 = 3448,
            MAP_MAP088_11 = 3449,
            MAP_MAP088_24 = 3450,
            MAP_MAP088_26 = 3451,
            MAP_MAP088_27 = 3452,
            MAP_MAP088_29 = 3453,
            MAP_MAP088_30 = 3454,
            MAP_MAP088_32 = 3455,
            MAP_MAP088_33 = 3456,
            MAP_MAP088_35 = 3457,
            MAP_MAP088_36 = 3458,
            MAP_MAP089_6 = 3461,
            MAP_MAP089_7 = 3462,
            MAP_MAP089_10 = 3463,
            MAP_MAP089_11 = 3464,
            MAP_MAP089_14 = 3465,
            MAP_MAP089_15 = 3466,
            MAP_MAP089_18 = 3467,
            MAP_MAP089_19 = 3468,
            MAP_MAP089_22 = 3469,
            MAP_MAP089_23 = 3470,
            MAP_MAP089_26 = 3471,
            MAP_MAP089_27 = 3472,
            MAP_MAP089_30 = 3473,
            MAP_MAP089_31 = 3474,
            MAP_MAP089_38 = 3475,
            MAP_MAP089_35 = 3476,
            MAP_MAP089_34 = 3477,
            MAP_MAP089_39 = 3478,
            MAP_MAP089_42 = 3479,
            MAP_MAP089_43 = 3480,
            MAP_MAP090_6 = 3483,
            MAP_MAP090_7 = 3484,
            MAP_MAP090_10 = 3485,
            MAP_MAP090_11 = 3486,
            MAP_MAP090_14 = 3487,
            MAP_MAP090_15 = 3488,
            MAP_MAP090_18 = 3489,
            MAP_MAP090_19 = 3490,
            MAP_MAP090_38 = 3491,
            MAP_MAP090_23 = 3492,
            MAP_MAP090_26 = 3493,
            MAP_MAP090_27 = 3494,
            MAP_MAP090_30 = 3495,
            MAP_MAP090_31 = 3496,
            MAP_MAP090_34 = 3497,
            MAP_MAP090_35 = 3498,
            MAP_MAP090_22 = 3499,
            MAP_MAP090_39 = 3500,
            MAP_MAP090_42 = 3501,
            MAP_MAP090_43 = 3502,
            MAP_MAP091_8 = 3505,
            MAP_MAP091_9 = 3506,
            MAP_MAP091_10 = 3507,
            MAP_MAP091_13 = 3508,
            MAP_MAP091_14 = 3509,
            MAP_MAP091_17 = 3510,
            MAP_MAP091_18 = 3511,
            MAP_MAP091_21 = 3512,
            MAP_MAP091_22 = 3513,
            MAP_MAP092_30 = 3516,
            MAP_MAP092_9 = 3517,
            MAP_MAP092_12 = 3518,
            MAP_MAP092_13 = 3519,
            MAP_MAP092_38 = 3520,
            MAP_MAP092_39 = 3521,
            MAP_MAP092_42 = 3522,
            MAP_MAP092_43 = 3523,
            MAP_MAP092_8 = 3524,
            MAP_MAP092_31 = 3525,
            MAP_MAP092_34 = 3526,
            MAP_MAP092_35 = 3527,
            MAP_MAP092_16 = 3528,
            MAP_MAP092_17 = 3529,
            MAP_MAP092_20 = 3530,
            MAP_MAP092_21 = 3531,
            MAP_MAP093_6 = 3534,
            MAP_MAP093_7 = 3535,
            MAP_MAP093_10 = 3536,
            MAP_MAP093_11 = 3537,
            MAP_MAP093_14 = 3538,
            MAP_MAP093_15 = 3539,
            MAP_MAP093_18 = 3540,
            MAP_MAP093_19 = 3541,
            MAP_MAP094_7 = 3544,
            MAP_MAP094_8 = 3545,
            MAP_MAP094_16 = 3546,
            MAP_MAP094_17 = 3547,
            MAP_MAP095_36 = 3549,
            MAP_MAP095_8 = 3550,
            MAP_MAP095_32 = 3551,
            MAP_MAP095_12 = 3552,
            MAP_MAP095_15 = 3553,
            MAP_MAP095_16 = 3554,
            MAP_MAP095_11 = 3555,
            MAP_MAP095_20 = 3556,
            MAP_MAP095_28 = 3557,
            MAP_MAP095_29 = 3558,
            MAP_MAP095_19 = 3559,
            MAP_MAP095_33 = 3560,
            MAP_MAP095_7 = 3561,
            MAP_MAP095_41 = 3562,
            MAP_MAP095_40 = 3563,
            MAP_MAP095_37 = 3564,
            MAP_MAP096_12 = 3566,
            MAP_MAP096_8 = 3567,
            MAP_MAP096_7 = 3568,
            MAP_MAP096_13 = 3569,
            MAP_MAP096_20 = 3570,
            MAP_MAP096_17 = 3571,
            MAP_MAP096_16 = 3572,
            MAP_MAP096_21 = 3573,
            MAP_MAP097_7 = 3575,
            MAP_MAP097_8 = 3576,
            MAP_MAP097_25 = 3577,
            MAP_MAP097_12 = 3578,
            MAP_MAP097_29 = 3579,
            MAP_MAP097_16 = 3580,
            MAP_MAP097_41 = 3581,
            MAP_MAP097_19 = 3582,
            MAP_MAP097_37 = 3583,
            MAP_MAP097_22 = 3584,
            MAP_MAP097_11 = 3585,
            MAP_MAP097_26 = 3586,
            MAP_MAP097_15 = 3587,
            MAP_MAP097_30 = 3588,
            MAP_MAP097_33 = 3589,
            MAP_MAP097_34 = 3590,
            MAP_MAP097_18 = 3591,
            MAP_MAP097_38 = 3592,
            MAP_MAP097_21 = 3593,
            MAP_MAP097_42 = 3594,
            MAP_MAP098_6 = 3596,
            MAP_MAP098_7 = 3597,
            MAP_MAP098_10 = 3598,
            MAP_MAP098_11 = 3599,
            MAP_MAP098_14 = 3600,
            MAP_MAP098_15 = 3601,
            MAP_MAP098_38 = 3602,
            MAP_MAP098_19 = 3603,
            MAP_MAP098_42 = 3604,
            MAP_MAP098_23 = 3605,
            MAP_MAP098_26 = 3606,
            MAP_MAP098_27 = 3607,
            MAP_MAP098_34 = 3608,
            MAP_MAP098_31 = 3609,
            MAP_MAP098_22 = 3610,
            MAP_MAP098_35 = 3611,
            MAP_MAP098_18 = 3612,
            MAP_MAP098_39 = 3613,
            MAP_MAP098_30 = 3614,
            MAP_MAP098_43 = 3615,
            MAP_MAP099_6 = 3617,
            MAP_MAP099_7 = 3618,
            MAP_MAP099_10 = 3619,
            MAP_MAP099_11 = 3620,
            MAP_MAP099_14 = 3621,
            MAP_MAP099_15 = 3622,
            MAP_MAP099_18 = 3623,
            MAP_MAP099_19 = 3624,
            MAP_MAP099_22 = 3625,
            MAP_MAP099_23 = 3626,
            MAP_MAP099_26 = 3627,
            MAP_MAP099_27 = 3628,
            MAP_MAP099_30 = 3629,
            MAP_MAP099_31 = 3630,
            MAP_MAP099_34 = 3631,
            MAP_MAP099_35 = 3632,
            MAP_MAP099_38 = 3633,
            MAP_MAP099_39 = 3634,
            MAP_MAP099_42 = 3635,
            MAP_MAP099_43 = 3636,
            MAP_MAP100_7 = 3638,
            MAP_MAP100_8 = 3639,
            MAP_MAP101_9 = 3642,
            MAP_MAP101_10 = 3643,
            MAP_MAP102_7 = 3645,
            MAP_MAP102_8 = 3646,
            MAP_MAP103_9 = 3648,
            MAP_MAP103_10 = 3649,
            MAP_MAP103_11 = 3650,
            MAP_MAP103_13 = 3651,
            MAP_MAP103_14 = 3652,
            MAP_MAP103_19 = 3653,
            MAP_MAP103_17 = 3654,
            MAP_MAP103_16 = 3655,
            MAP_MAP103_20 = 3656,
            MAP_MAP103_23 = 3657,
            MAP_MAP103_24 = 3658,
            MAP_MAP103_27 = 3659,
            MAP_MAP103_28 = 3660,
            MAP_MAP103_31 = 3661,
            MAP_MAP103_32 = 3662,
            MAP_MAP103_39 = 3663,
            MAP_MAP103_36 = 3664,
            MAP_MAP103_35 = 3665,
            MAP_MAP103_40 = 3666,
            MAP_MAP104_8 = 3669,
            MAP_MAP104_9 = 3670,
            MAP_MAP104_10 = 3671,
            MAP_MAP105_6 = 3674,
            MAP_MAP105_15 = 3675,
            MAP_MAP105_14 = 3676,
            MAP_MAP105_7 = 3677,
            MAP_MAP105_17 = 3678,
            MAP_MAP105_19 = 3679,
            MAP_MAP105_21 = 3680,
            MAP_MAP105_23 = 3681,
            MAP_MAP106_6 = 3683,
            MAP_MAP106_7 = 3684,
            MAP_MAP106_9 = 3685,
            MAP_MAP106_11 = 3686,
            MAP_MAP106_13 = 3687,
            MAP_MAP106_15 = 3688,
            MAP_MAP106_17 = 3689,
            MAP_MAP107_6 = 3691,
            MAP_MAP107_7 = 3692,
            MAP_MAP107_9 = 3693,
            MAP_MAP107_11 = 3694,
            MAP_MAP107_13 = 3695,
            MAP_MAP107_15 = 3696,
            MAP_MAP107_17 = 3697,
            MAP_MAP108_6 = 3699,
            MAP_MAP108_7 = 3700,
            MAP_MAP108_9 = 3701,
            MAP_MAP108_11 = 3702,
            MAP_MAP108_13 = 3703,
            MAP_MAP108_15 = 3704,
            MAP_MAP108_17 = 3705,
            MAP_MAP109_6 = 3707,
            MAP_MAP109_7 = 3708,
            MAP_MAP109_9 = 3709,
            MAP_MAP109_11 = 3710,
            MAP_MAP109_13 = 3711,
            MAP_MAP109_15 = 3712,
            MAP_MAP109_17 = 3713,
            MAP_MAP110_6 = 3715,
            MAP_MAP110_7 = 3716,
            MAP_MAP110_9 = 3717,
            MAP_MAP110_11 = 3718,
            MAP_MAP110_13 = 3719,
            MAP_MAP110_15 = 3720,
            MAP_MAP110_17 = 3721,
            MAP_MAP111_8 = 3723,
            MAP_MAP111_9 = 3724,
            MAP_MAP111_11 = 3725,
            MAP_MAP111_13 = 3726,
            MAP_MAP111_15 = 3727,
            MAP_MAP111_17 = 3728,
            MAP_MAP111_19 = 3729,
            MAP_MAP112_6 = 3731,
            MAP_MAP112_7 = 3732,
            MAP_MAP112_8 = 3733,
            MAP_MAP112_10 = 3734,
            MAP_MAP112_12 = 3735,
            MAP_MAP112_14 = 3736,
            MAP_MAP112_16 = 3737,
            MAP_MAP113_6 = 3739,
            MAP_MAP113_7 = 3740,
            MAP_MAP113_9 = 3741,
            MAP_MAP113_11 = 3742,
            MAP_MAP113_13 = 3743,
            MAP_MAP113_15 = 3744,
            MAP_MAP113_17 = 3745,
            MAP_MAP114_6 = 3747,
            MAP_MAP114_7 = 3748,
            MAP_MAP114_9 = 3749,
            MAP_MAP114_11 = 3750,
            MAP_MAP114_13 = 3751,
            MAP_MAP114_15 = 3752,
            MAP_MAP114_17 = 3753,
            MAP_MAP115_6 = 3755,
            MAP_MAP115_7 = 3756,
            MAP_MAP115_10 = 3757,
            MAP_MAP115_11 = 3758,
            MAP_MAP115_26 = 3759,
            MAP_MAP115_15 = 3760,
            MAP_MAP115_38 = 3761,
            MAP_MAP115_19 = 3762,
            MAP_MAP115_34 = 3763,
            MAP_MAP115_23 = 3764,
            MAP_MAP115_14 = 3765,
            MAP_MAP115_27 = 3766,
            MAP_MAP115_30 = 3767,
            MAP_MAP115_31 = 3768,
            MAP_MAP115_42 = 3769,
            MAP_MAP115_35 = 3770,
            MAP_MAP115_18 = 3771,
            MAP_MAP115_39 = 3772,
            MAP_MAP115_22 = 3773,
            MAP_MAP115_43 = 3774,
            MAP_MAP116_6 = 3777,
            MAP_MAP116_7 = 3778,
            MAP_MAP117_6 = 3780,
            MAP_MAP117_7 = 3781,
            MAP_MAP119_6 = 3783,
            MAP_MAP118_7 = 3784,
            MAP_MAP118_6 = 3786,
            MAP_MAP119_7 = 3787,
            MAP_MAP125_6 = 3789,
            MAP_MAP125_7 = 3790,
            MAP_MAP011_22_duplicate = 3821,
            MAP_MAP011_28_duplicate = 3822,
            MAP_MAP011_53_duplicate = 3827,
        }
    }
}
